unit Display;
{ TDisplay 1.0 (c) 1996 by Gerry Skolnik (skolnik@kapsch.co.at)

  This component emulates a 1 digit electronic display in three styles,
  7-segment, 16-segment and 5x7 dot matrix.

  This release is for Delphi 2. To install this component into the Delphi 1 library,
  perform these two steps:

  o) Change the uses statement below from 'uses Windows, ...'
                                       to 'uses WinTypes, WinProcs... '

  o) Rename the file disp16.dcr to display.dcr.


  TDisplay is Freeware. You may use it freely at your own risk in any
  kind of environment. The component itself must be distributed along with
  the source code. If you use this component as part of any project, you
  of course do _not_ have to include this source code...

  A little advice: this component is freely resizeable, thus experimenting with Margin,
  SegWidth, PenOnColor etc. is definitely encouraged to provide the best appearance.
  For small sizes, PenOnColor should be the same as OnColor,  PenOffColor = OffColor.

  The properties:

  Digit          ...  the character to be displayed (hexadecimal for 7-segment,
                      alphanumeric for 16-segment (uppercase plus a few special chars,
                      lowercase chars are displayed uppercase),
                      full ASCII for 5x7 matrix

  OnColor        ...  Color of lit segment

  PenOnColor     ...  Color of pen to draw lit segment

  OffColor       ...  Color of dark segment

  PenOffColor    ...  Color of pen to draw dark segment

  BackColor      ...  Background color of display

  SegWidth       ...  Width (7-segment, 16-segment) or diameter (5x7 matrix) of segment

  Margin         ...  Distance between border of display and outer segments

  DigitStyle     ...  ds7Seg (7-Segment), ds16Seg (16-Segment), ds5x7Dot (5x7 square dots),
                      ds5x7Bulb (5x7 circle dots)

  The characters are made up of bitmasks; if the segment's according bit is set -
  surprise - it will be lit, otherwise it will stay dark. The segments have been
  coded the following way:

  ds7Seg:  0 ... top
           1 ... center
           2 ... bottom
           3 ... upper left
           4 ... upper right
           5 ... lower left
           6 ... lower right

           Where have the decimal dots gone to? Nowhere, they've never been there as I
           didn't require them in the project this component was developped for. I might
           include them in another release though, if by popular demand. Or feel free to
           do it yourself. Just give me a call if you do, ok?

  ds5x7##: each character made up of 7 bytes, which are thus stored in an array of 896
           elements (128 x 7). The first byte describes the top most row, the seventh
           byte the bottom most. Each byte carries 5 significant bits, with bit 0 being
           the right most dot and bit 5 being the left most.

           The character map was taken from a Hewlett Packard data sheet for their
           HDSP 2470.

  ds16Seg: 16 segment bits conveniently fitting, each character is made up of one owrd.
           The segments have been coded the following way:

           0 ... top left
           1 ... top right
           2 ... upper left
           3 ... north west
           4 ... upper center (vertical)
           5 ... north east
           6 ... upper right
           7 ... left center (horizontal)
           8 ... right center (horizontal)
           9 ... lower left
           10 .. south west
           11 .. lower center (vertical)
           12 .. south east
           13 .. lower right
           14 .. bottom left
           15 .. bottom right

}
interface

uses
  Windows, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls;

type
  TDigitStyle = (ds7Seg, ds16Seg, ds5x7Dot, ds5x7Bulb);

  TDisplay = class(TGraphicControl)
  private
    FDigit      : char;
    FOnColor    : TColor;
    FOffColor   : TColor;
    FBackColor  : TColor;
    FSegWidth   : integer;
    FMargin     : integer;
    FDigitStyle : TDigitStyle;
    FPenOnColor : TColor;
    FPenOffColor: TColor;
    UpLeft      : array[0..34] of TPoint;
    SegHorLen   : integer;
    SegVerLen   : integer;
    SegHalf     : integer;
    OldWidth    : integer;
    OldHeight   : integer;
    procedure SetDigit(WhatDigit: char);
    procedure SetOnColor(WhatColor: TColor);
    procedure SetOffColor(WhatColor: TColor);
    procedure SetBackColor(WhatColor: TColor);
    procedure SetSegWidth(WhatWidth: integer);
    procedure SetMargin(WhatWidth: integer);
    procedure SetDigitStyle(WhatStyle: TDigitStyle);
    procedure SetPenOnColor(WhatColor: TColor);
    procedure SetPenOffColor(WhatColor: TColor);
    procedure Hexagon(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
    procedure DrawDot(WhereTo: TBitmap; x, y: integer; OnOff: boolean);
    procedure DrawBulb(WhereTo: TBitmap; x, y: integer; OnOff: boolean);
    procedure DrawTopSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
    procedure DrawBottomSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
    procedure DrawLeftSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
    procedure DrawRightSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
    procedure DrawNorthWest(WhereTo: TBitmap; x1, y1, x2, y2: integer; OnOff: boolean);
    procedure DrawNorthEast(WhereTo: TBitmap; x1, y1, x2, y2: integer; OnOff: boolean);
    procedure DrawDigit(WhereTo: TBitmap);
    procedure SetMeasures;
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Digit: char read FDigit write SetDigit default #0;
    property OnColor: TColor read FOnColor write SetOnColor default clRed;
    property OffColor: TColor read FOffColor write SetOffColor default clGray;
    property BackColor: TColor read FBackColor write SetBackColor default clBlack;
    property SegWidth: integer read FSegWidth write SetSegWidth default 15;
    property Margin: integer read FMargin write SetMargin default 10;
    property DigitStyle: TDigitStyle read FDigitStyle write SetDigitStyle default ds7Seg;
    property PenOnColor: TColor read FPenOnColor write SetPenOnColor default clGray;
    property PenOffColor: TColor read FPenOffColor write SetPenOffColor default clGray;
  end;

procedure Register;

implementation

const
  SegState: array[0..127] of word =   { all ASCII chars, display '-' if unknown }
              ($0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $007d, $0050, $0037, $0057, $005a, $004f, $006f, $0051,   { 0 1 2 3 4 5 6 7 }
               $007f, $005f, $0002, $0002, $0002, $0002, $0002, $0002,   { 8 9             }
               $0002, $007b, $006e, $002d, $0076, $002f, $002b, $0002,   { A B C D E F G   }
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $007b, $006e, $002d, $0076, $002f, $002b, $0002,   { a b c d e f g   }
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002,
               $0002, $0002, $0002, $0002, $0002, $0002, $0002, $0002);

  DotState: array[0..895] of byte = { all ASCII according to Hewlett Packard HDSP 2470 }
              ($02, $06, $0e, $1e, $0e, $06, $02,     { NUL }
               $04, $00, $04, $08, $11, $11, $0e,     { SOH }
               $1f, $00, $11, $0a, $04, $0a, $11,     { STX }
               $1f, $00, $11, $19, $15, $13, $11,     { ETX }
               $1f, $00, $16, $19, $11, $11, $11,     { EOT }
               $00, $00, $0d, $12, $12, $12, $0d,     { ENQ }
               $1c, $12, $12, $16, $11, $16, $10,     { ACK }
               $06, $08, $04, $0e, $11, $11, $0e,     { BEL }
               $00, $00, $00, $04, $0a, $11, $1f,     { BS }
               $00, $10, $1c, $12, $12, $02, $01,     { HT }
               $0e, $11, $11, $1f, $11, $11, $0e,     { LF }
               $00, $10, $08, $04, $0a, $11, $11,     { VT }
               $00, $00, $09, $09, $09, $0e, $10,     { FF }
               $00, $01, $0e, $1a, $0a, $0a, $0a,     { CR }
               $00, $00, $0f, $12, $12, $12, $0c,     { SO }
               $1f, $08, $04, $02, $04, $08, $1f,     { SI }
               $00, $00, $01, $0e, $14, $04, $04,     { DLE }
               $00, $04, $0e, $15, $15, $0e, $04,     { DC1 }
               $0e, $11, $11, $11, $11, $0a, $1b,     { DC2 }
               $04, $00, $0e, $11, $1f, $11, $11,     { DC3 }
               $04, $00, $0e, $12, $12, $12, $0d,     { DC4 }
               $0a, $00, $0e, $11, $1f, $11, $11,     { NAK }
               $0a, $00, $0e, $12, $12, $12, $0d,     { SYN }
               $0a, $0e, $11, $11, $11, $11, $0e,     { ETB }
               $0a, $00, $0e, $11, $11, $11, $0e,     { CAN }
               $0a, $00, $11, $11, $11, $11, $0e,     { EM }
               $00, $0a, $00, $11, $11, $11, $0e,     { SUB }
               $00, $04, $02, $1f, $02, $04, $00,     { ESC }
               $00, $0f, $08, $08, $08, $18, $08,     { FS }
               $0c, $12, $04, $08, $1e, $00, $00,     { GS }
               $06, $09, $08, $1c, $08, $08, $1f,     { RS }
               $11, $0a, $04, $04, $0e, $04, $04,     { US }
               $00, $00, $00, $00, $00, $00, $00,     { Space }
               $08, $08, $08, $08, $08, $00, $08,     { ! }
               $0a, $0a, $00, $00, $00, $00, $00,     { " }
               $0a, $0a, $1f, $0a, $1f, $0a, $0a,     { # }
               $04, $0f, $14, $0e, $05, $1e, $04,     { $ }
               $18, $19, $02, $04, $08, $13, $03,     { % }
               $08, $14, $14, $08, $15, $12, $0d,     { & }
               $0c, $0c, $04, $08, $00, $00, $00,     { ' }
               $02, $04, $04, $04, $04, $04, $02,     { ( }
               $08, $04, $04, $04, $04, $04, $08,     { ) }
               $00, $0a, $04, $1f, $04, $0a, $00,     { * }
               $00, $04, $04, $1f, $04, $04, $00,     { + }
               $00, $00, $00, $0c, $0c, $04, $08,     { , }
               $00, $00, $00, $1f, $00, $00, $00,     { - }
               $00, $00, $00, $00, $0c, $0c, $00,     { . }
               $00, $01, $02, $04, $08, $10, $00,     { / }
               $0e, $11, $13, $15, $19, $11, $0e,     { 0 }
               $04, $0c, $04, $04, $04, $04, $0e,     { 1 }
               $0e, $11, $01, $06, $08, $10, $1f,     { 2 }
               $0e, $11, $01, $06, $01, $11, $0e,     { 3 }
               $02, $06, $0a, $12, $1f, $02, $02,     { 4 }
               $1f, $10, $1e, $01, $01, $11, $0e,     { 5 }
               $06, $08, $10, $1e, $11, $11, $0e,     { 6 }
               $1f, $01, $02, $04, $08, $08, $08,     { 7 }
               $0e, $11, $11, $0e, $11, $11, $0e,     { 8 }
               $0e, $11, $11, $0f, $01, $02, $0c,     { 9 }
               $00, $0c, $0c, $00, $0c, $0c, $00,     { : }
               $0c, $0c, $00, $0c, $0c, $04, $08,     { ; }
               $01, $02, $04, $08, $04, $02, $01,     { < }
               $00, $00, $1f, $00, $1f, $00, $00,     { = }
               $10, $08, $04, $02, $04, $08, $10,     { > }
               $0e, $11, $01, $02, $04, $00, $04,     { ? }
               $0e, $11, $17, $15, $17, $10, $0e,     { @ }
               $0e, $11, $11, $1f, $11, $11, $11,     { A }
               $1e, $11, $11, $1e, $11, $11, $1e,     { B }
               $0e, $11, $10, $10, $10, $11, $0e,     { C }
               $1e, $11, $11, $11, $11, $11, $1e,     { D }
               $1f, $10, $10, $1e, $10, $10, $1f,     { E }
               $1f, $10, $10, $1e, $10, $10, $10,     { F }
               $0e, $11, $10, $10, $13, $11, $0f,     { G }
               $11, $11, $11, $1f, $11, $11, $11,     { H }
               $0e, $04, $04, $04, $04, $04, $0e,     { I }
               $01, $01, $01, $01, $01, $11, $0e,     { J }
               $11, $12, $14, $18, $14, $12, $11,     { K }
               $10, $10, $10, $10, $10, $10, $1f,     { L }
               $11, $1b, $15, $15, $11, $11, $11,     { M }
               $11, $11, $19, $15, $13, $11, $11,     { N }
               $0e, $11, $11, $11, $11, $11, $0e,     { O }
               $1e, $11, $11, $1e, $10, $10, $10,     { P }
               $0e, $11, $11, $11, $15, $12, $0d,     { Q }
               $1e, $11, $11, $1e, $14, $12, $11,     { R }
               $0e, $11, $10, $0e, $01, $11, $0e,     { S }
               $1f, $04, $04, $04, $04, $04, $04,     { T }
               $11, $11, $11, $11, $11, $11, $0e,     { U }
               $11, $11, $11, $0a, $0a, $04, $04,     { V }
               $11, $11, $11, $15, $15, $1b, $11,     { W }
               $11, $11, $0a, $04, $0a, $11, $11,     { X }
               $11, $11, $0a, $04, $04, $04, $04,     { Y }
               $1f, $01, $02, $04, $08, $10, $1f,     { Z }
               $07, $04, $04, $04, $04, $04, $07,     { [ }
               $00, $10, $08, $04, $02, $01, $00,     { \ }
               $1c, $04, $04, $04, $04, $04, $1c,     { ] }
               $04, $0e, $15, $04, $04, $04, $04,     { ^ }
               $00, $00, $00, $00, $00, $00, $1f,     { _ }
               $0c, $0c, $08, $04, $00, $00, $00,     { ` }
               $00, $00, $0e, $12, $12, $12, $0d,     { a }
               $10, $10, $16, $19, $11, $11, $1e,     { b }
               $00, $00, $0e, $10, $10, $11, $0e,     { c }
               $01, $01, $0d, $13, $11, $11, $0f,     { d }
               $00, $00, $0e, $11, $1e, $10, $0e,     { e }
               $04, $0a, $08, $1c, $08, $08, $08,     { f }
               $00, $00, $0f, $11, $0f, $01, $06,     { g }
               $10, $10, $16, $19, $11, $11, $11,     { h }
               $04, $00, $0c, $04, $04, $04, $0e,     { i }
               $02, $00, $06, $02, $02, $12, $0c,     { j }
               $08, $08, $09, $0a, $0c, $0a, $09,     { k }
               $0c, $04, $04, $04, $04, $04, $0e,     { l }
               $00, $00, $0a, $15, $15, $11, $11,     { m }
               $00, $00, $16, $19, $11, $11, $11,     { n }
               $00, $00, $0e, $11, $11, $11, $0e,     { o }
               $00, $00, $1e, $11, $19, $16, $10,     { p }
               $00, $00, $0e, $12, $16, $0a, $03,     { q }
               $00, $00, $0b, $0c, $08, $08, $08,     { r }
               $00, $00, $0e, $10, $0e, $01, $1e,     { s }
               $00, $08, $1c, $08, $08, $0a, $04,     { t }
               $00, $00, $11, $11, $11, $13, $0d,     { u }
               $00, $00, $11, $11, $11, $0a, $04,     { v }
               $00, $00, $11, $11, $15, $15, $0a,     { w }
               $00, $00, $11, $0a, $04, $0a, $11,     { x }
               $00, $00, $11, $0a, $04, $04, $08,     { y }
               $00, $00, $1f, $02, $04, $08, $1f,     { z }
               $02, $04, $04, $08, $04, $04, $02,     (* { *)
               $04, $04, $04, $00, $04, $04, $04,     { | }
               $08, $04, $04, $02, $04, $04, $08,     (* } *)
               $00, $00, $08, $15, $02, $00, $00,     { ~ }
               $0a, $15, $0a, $15, $0a, $15, $0a);    { DEL }

  SixState: array[0..127] of word =   { ASCII according to HPDL 1414, display '-' if unknown }
              ($0180, $0180, $0180, $0180, $0180, $0180, $0180, $0180,
               $0180, $0180, $0180, $0180, $0180, $0180, $0180, $0180,
               $0180, $0180, $0180, $0180, $0180, $0180, $0180, $0180,
               $0180, $0180, $0180, $0180, $0180, $0180, $0180, $0180,
               $0000, $0180, $0014, $e9d0, $e997, $adb5, $f299, $0020,   { sp ... }
               $1020, $0408, $1db8, $0990, $0400, $0180, $0180, $0420,
               $4a15, $0810, $4291, $4891, $0894, $4885, $4a85, $0811,   { 0 1 2 3 4 5 6 7 }
               $4a95, $4895, $4001, $0401, $c420, $c180, $d008, $0943,   { 8 9             }
               $eac3, $23c7, $e953, $c207, $e853, $c287, $0287, $e307,   { @ A B C D E F G }
               $23c4, $c813, $4a13, $12a4, $c204, $226c, $324c, $e247,
               $03c7, $f247, $13c7, $e187, $0813, $e244, $0624, $3644,
               $1428, $0828, $c423, $4205, $1008, $a042, $1400, $c000,
               $0180, $23c7, $e953, $c207, $e853, $c287, $0287, $e307,   {   a b c d e f g }
               $23c4, $c813, $4a13, $12a4, $c204, $226c, $324c, $e247,
               $03c7, $f247, $13c7, $e187, $0813, $e244, $0624, $3644,
               $1428, $0828, $c423, $0180, $0180, $0180, $0180, $0180);

constructor TDisplay.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  Width := 120;
  Height := 200;
  FOnColor := clRed;
  FOffColor := clGray;
  FBackColor := clBlack;
  FSegWidth := 15;
  FMargin := 10;
  OldWidth := Width;
  OldHeight := Height;
  SetMeasures;
end;

procedure TDisplay.SetDigit(WhatDigit: char);
begin
  if WhatDigit <> FDigit then
  begin
    FDigit := WhatDigit;
    Refresh;
  end;
end;

procedure TDisplay.SetOnColor(WhatColor: TColor);
begin
  if WhatColor <> FOnColor then
  begin
    FOnColor := WhatColor;
    Refresh;
  end;
end;

procedure TDisplay.SetPenOnColor(WhatColor: TColor);
begin
  if WhatColor <> FPenOnColor then
  begin
    FPenOnColor := WhatColor;
    Refresh;
  end;
end;

procedure TDisplay.SetOffColor(WhatColor: TColor);
begin
  if WhatColor <> FOffColor then
  begin
    FOffColor := WhatColor;
    Refresh;
  end;
end;

procedure TDisplay.SetPenOffColor(WhatColor: TColor);
begin
  if WhatColor <> FPenOffColor then
  begin
    FPenOffColor := WhatColor;
    Refresh;
  end;
end;

procedure TDisplay.SetBackColor(WhatColor: TColor);
begin
  if WhatColor <> FBackColor then
  begin
    FBackColor := WhatColor;
    Refresh;
  end;
end;

procedure TDisplay.SetSegWidth(WhatWidth: integer);
begin
  if WhatWidth <> FSegWidth then
  begin
    FSegWidth := WhatWidth;
    SetMeasures;
    Refresh;
  end;
end;

procedure TDisplay.SetMargin(WhatWidth: integer);
begin
  if WhatWidth <> FMargin then
  begin
    FMargin := WhatWidth;
    SetMeasures;
    Refresh;
  end;
end;

procedure TDisplay.SetDigitStyle(WhatStyle: TDigitStyle);
begin
  if WhatStyle <> FDigitStyle then
  begin
    FDigitStyle := WhatStyle;
    SetMeasures;
    Refresh;
  end;
end;

procedure TDisplay.Hexagon(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  if h > w then { vertical, w is SegWidth }
    WhereTo.Canvas.Polygon([Point(x, y + SegHalf), Point(x + SegHalf, y),
                            Point(x + w, y + SegHalf), Point(x + w, y + h - SegHalf),
                            Point(x + SegHalf, y + h), Point(x, y + h - SegHalf)])
  else          { horizontal, h is SegWidth }
    WhereTo.Canvas.Polygon([Point(x, y + SegHalf), Point(x + SegHalf, y),
                            Point(x + w - SegHalf, y), Point(x + w, y + SegHalf),
                            Point(x + w - SegHalf, y + h), Point(x + SegHalf, y + h)]);
end;

procedure TDisplay.DrawDot(WhereTo: TBitmap; x, y: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Rectangle(x, y, x + FSegWidth, y + FSegWidth);
end;

procedure TDisplay.DrawBulb(WhereTo: TBitmap; x, y: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Ellipse(x, y, x + FSegWidth, y + FSegWidth);
end;

procedure TDisplay.DrawTopSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x, y), Point(x + FSegWidth, y + h),
                          Point(x + w - FSegWidth, y + h), Point(x + w, y)]);
end;

procedure TDisplay.DrawBottomSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x, y + h), Point(x + FSegWidth, y),
                          Point(x + w - FSegWidth, y), Point(x + w, y + h)]);
end;

procedure TDisplay.DrawLeftSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x, y), Point(x + w, y + FSegWidth),
                          Point(x + w, y + h - FSegWidth), Point(x, y + h)]);
end;

procedure TDisplay.DrawRightSegment(WhereTo: TBitmap; x, y, w, h: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x + w, y), Point(x, y + FSegWidth),
                          Point(x, y + h - FSegWidth), Point(x + w, y + h)]);
end;

procedure TDisplay.DrawNorthWest(WhereTo: TBitmap; x1, y1, x2, y2: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x1, y1), Point(x1 + SegHalf, y1),
                          Point(x2, y2 - FSegWidth), Point(x2, y2),
                          Point(x2 - SegHalf, y2), Point(x1, y1 + FSegWidth)]);
end;

procedure TDisplay.DrawNorthEast(WhereTo: TBitmap; x1, y1, x2, y2: integer; OnOff: boolean);
begin
  if OnOff then
  begin
    WhereTo.Canvas.Brush.Color := FOnColor;
    WhereTo.Canvas.Pen.Color := FPenOnColor;
  end
  else
  begin
    WhereTo.Canvas.Brush.Color := FOffColor;
    WhereTo.Canvas.Pen.Color := FPenOffColor;
  end;
  WhereTo.Canvas.Polygon([Point(x1, y1), Point(x1, y1 - FSegWidth),
                          Point(x2 - SegHalf, y2), Point(x2, y2),
                          Point(x2, y2 + FSegWidth), Point(x1 + SegHalf, y1)]);
end;

procedure TDisplay.SetMeasures;
var
  i, j: integer;
begin
  case FDigitStyle of
    ds7Seg:
      begin
        SegHalf := FSegWidth div 2;
        SegHorLen := Width - FMargin * 2 - FSegWidth;
        SegVerLen := Height div 2 - FSegWidth - SegHalf - FMargin;
        UpLeft[0] := Point(FMargin + SegHalf, FMargin);                         { 3 hor }
        UpLeft[1] := Point(FMargin + SegHalf, Height div 2 - SegHalf);
        UpLeft[2] := Point(FMargin + SegHalf, Height - FMargin - FSegWidth);
        UpLeft[3] := Point(FMargin, FMargin + FSegWidth);                               { 4 ver }
        UpLeft[4] := Point(Width - FMargin - FSegWidth, FMargin + FSegWidth);
        UpLeft[5] := Point(FMargin, Height div 2 + SegHalf);
        UpLeft[6] := Point(Width - FMargin - FSegWidth, Height div 2 + SegHalf);
      end;
    ds5x7Dot, ds5x7Bulb:
      begin
        i := (Width - FMargin * 2 - FSegWidth * 5) div 4;
        j := (Height - FMargin * 2 - FSegWidth * 7) div 6;
        UpLeft[0] := Point(FMargin, FMargin);
        UpLeft[1] := Point(FMargin + FSegWidth + i, FMargin);
        UpLeft[2] := Point(FMargin + (FSegWidth + i) * 2, FMargin);
        UpLeft[3] := Point(FMargin + (FSegWidth + i) * 3, FMargin);
        UpLeft[4] := Point(FMargin + (FSegWidth + i) * 4, FMargin);
        UpLeft[5] := Point(FMargin, FMargin + FSegWidth + j);
        UpLeft[6] := Point(FMargin + FSegWidth + i, FMargin + FSegWidth + j);
        UpLeft[7] := Point(FMargin + (FSegWidth + i) * 2, FMargin + FSegWidth + j);
        UpLeft[8] := Point(FMargin + (FSegWidth + i) * 3, FMargin + FSegWidth + j);
        UpLeft[9] := Point(FMargin + (FSegWidth + i) * 4, FMargin + FSegWidth + j);
        UpLeft[10] := Point(FMargin, FMargin + (FSegWidth + j) * 2);
        UpLeft[11] := Point(FMargin + FSegWidth + i, FMargin + (FSegWidth + j) * 2);
        UpLeft[12] := Point(FMargin + (FSegWidth + i) * 2, FMargin + (FSegWidth + j) * 2);
        UpLeft[13] := Point(FMargin + (FSegWidth + i) * 3, FMargin + (FSegWidth + j) * 2);
        UpLeft[14] := Point(FMargin + (FSegWidth + i) * 4, FMargin + (FSegWidth + j) * 2);
        UpLeft[15] := Point(FMargin, FMargin + (FSegWidth + j) * 3);
        UpLeft[16] := Point(FMargin + FSegWidth + i, FMargin + (FSegWidth + j) * 3);
        UpLeft[17] := Point(FMargin + (FSegWidth + i) * 2, FMargin + (FSegWidth + j) * 3);
        UpLeft[18] := Point(FMargin + (FSegWidth + i) * 3, FMargin + (FSegWidth + j) * 3);
        UpLeft[19] := Point(FMargin + (FSegWidth + i) * 4, FMargin + (FSegWidth + j) * 3);
        UpLeft[20] := Point(FMargin, FMargin + (FSegWidth + j) * 4);
        UpLeft[21] := Point(FMargin + FSegWidth + i, FMargin + (FSegWidth + j) * 4);
        UpLeft[22] := Point(FMargin + (FSegWidth + i) * 2, FMargin + (FSegWidth + j) * 4);
        UpLeft[23] := Point(FMargin + (FSegWidth + i) * 3, FMargin + (FSegWidth + j) * 4);
        UpLeft[24] := Point(FMargin + (FSegWidth + i) * 4, FMargin + (FSegWidth + j) * 4);
        UpLeft[25] := Point(FMargin, FMargin + (FSegWidth + j) * 5);
        UpLeft[26] := Point(FMargin + FSegWidth + i, FMargin + (FSegWidth + j) * 5);
        UpLeft[27] := Point(FMargin + (FSegWidth + i) * 2, FMargin + (FSegWidth + j) * 5);
        UpLeft[28] := Point(FMargin + (FSegWidth + i) * 3, FMargin + (FSegWidth + j) * 5);
        UpLeft[29] := Point(FMargin + (FSegWidth + i) * 4, FMargin + (FSegWidth + j) * 5);
        UpLeft[30] := Point(FMargin, FMargin + (FSegWidth + j) * 6);
        UpLeft[31] := Point(FMargin + FSegWidth + i, FMargin + (FSegWidth + j) * 6);
        UpLeft[32] := Point(FMargin + (FSegWidth + i) * 2, FMargin + (FSegWidth + j) * 6);
        UpLeft[33] := Point(FMargin + (FSegWidth + i) * 3, FMargin + (FSegWidth + j) * 6);
        UpLeft[34] := Point(FMargin + (FSegWidth + i) * 4, FMargin + (FSegWidth + j) * 6);
      end;
    ds16Seg:
      begin
        SegHalf := FSegWidth div 2;
        SegHorLen := Width div 2 - FMargin - FSegWidth;
        SegVerLen := Height div 2 - FSegWidth - FMargin;
        UpLeft[0] := Point(FMargin + SegHalf, FMargin);                            { top }
        UpLeft[1] := Point(Width div 2 + SegHalf, FMargin);
        UpLeft[14] := Point(FMargin + SegHalf, Height - FMargin - FSegWidth);      { bottom }
        UpLeft[15] := Point(Width div 2 + SegHalf, Height - FMargin - FSegWidth);
        UpLeft[2] := Point(FMargin, FMargin + SegHalf);                            { left }
        UpLeft[9] := Point(FMargin, Height div 2 + SegHalf);
        UpLeft[6] := Point(Width - FMargin - FSegWidth, FMargin + SegHalf);        { right }
        UpLeft[13] := Point(Width - FMargin - FSegWidth, Height div 2 + SegHalf);
        UpLeft[4] := Point(Width div 2  - SegHalf, FMargin + SegHalf);             { vertical }
        UpLeft[11] := Point(Width div 2 - SegHalf, Height div 2 + SegHalf);
        UpLeft[7] := Point(FMargin + SegHalf, Height div 2 - SegHalf);             { horizontal }
        UpLeft[8] := Point(Width div 2 + SegHalf, Height div 2 - SegHalf);

        UpLeft[3] := Point(FMargin + FSegWidth + SegHalf, FMargin + FSegWidth + SegHalf);  { NW }
        UpLeft[23] := Point(Width div 2 - FSegWidth, Height div 2 - FSegWidth);

        UpLeft[5] := Point(Width div 2 + FSegWidth, Height div 2 - FSegWidth);             { NO }
        UpLeft[25] := Point(Width - FMargin - FSegWidth - SegHalf,
                            FMargin + FSegWidth + SegHalf);

        UpLeft[10] := Point(FMargin + FSegWidth + SegHalf,                                 { SW }
                            Height - FMargin - FSegWidth - SegHalf);
        UpLeft[30] := Point(Width div 2 - FSegWidth, Height div 2 + FSegWidth);

        UpLeft[12] := Point(Width div 2 + FSegWidth, Height div 2 + FSegWidth);            { SO }
        UpLeft[32] := Point(Width - FMargin - FSegWidth - SegHalf,
                            Height - FMargin - FSegWidth - SegHalf);
      end;
  end;
end;

procedure TDisplay.DrawDigit(WhereTo: TBitmap);
var
  i : integer;
begin
  case FDigitStyle of
    ds7Seg:
      begin
        with WhereTo.Canvas do
        begin
          i := ord(FDigit);
          Hexagon(WhereTo, UpLeft[0].x, UpLeft[0].y, SegHorLen, FSegWidth, SegState[i] and $0001 <> 0);
          Hexagon(WhereTo, UpLeft[1].x, UpLeft[1].y, SegHorLen, FSegWidth, SegState[i] and $0002 <> 0);
          Hexagon(WhereTo, UpLeft[2].x, UpLeft[2].y, SegHorLen, FSegWidth, SegState[i] and $0004 <> 0);
          Hexagon(WhereTo, UpLeft[3].x, UpLeft[3].y, FSegWidth, SegVerLen, SegState[i] and $0008 <> 0);
          Hexagon(WhereTo, UpLeft[4].x, UpLeft[4].y, FSegWidth, SegVerLen, SegState[i] and $0010 <> 0);
          Hexagon(WhereTo, UpLeft[5].x, UpLeft[5].y, FSegWidth, SegVerLen, SegState[i] and $0020 <> 0);
          Hexagon(WhereTo, UpLeft[6].x, UpLeft[6].y, FSegWidth, SegVerLen, SegState[i] and $0040 <> 0);
        end;
      end;
    ds5x7Dot:
      begin
        with WhereTo.Canvas do
        begin
          i := ord(FDigit) * 7;
          DrawDot(WhereTo, UpLeft[0].x, UpLeft[0].y, DotState[i] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[1].x, UpLeft[1].y, DotState[i] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[2].x, UpLeft[2].y, DotState[i] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[3].x, UpLeft[3].y, DotState[i] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[4].x, UpLeft[4].y, DotState[i] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[5].x, UpLeft[5].y, DotState[i + 1] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[6].x, UpLeft[6].y, DotState[i + 1] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[7].x, UpLeft[7].y, DotState[i + 1] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[8].x, UpLeft[8].y, DotState[i + 1] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[9].x, UpLeft[9].y, DotState[i + 1] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[10].x, UpLeft[10].y, DotState[i + 2] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[11].x, UpLeft[11].y, DotState[i + 2] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[12].x, UpLeft[12].y, DotState[i + 2] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[13].x, UpLeft[13].y, DotState[i + 2] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[14].x, UpLeft[14].y, DotState[i + 2] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[15].x, UpLeft[15].y, DotState[i + 3] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[16].x, UpLeft[16].y, DotState[i + 3] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[17].x, UpLeft[17].y, DotState[i + 3] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[18].x, UpLeft[18].y, DotState[i + 3] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[19].x, UpLeft[19].y, DotState[i + 3] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[20].x, UpLeft[20].y, DotState[i + 4] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[21].x, UpLeft[21].y, DotState[i + 4] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[22].x, UpLeft[22].y, DotState[i + 4] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[23].x, UpLeft[23].y, DotState[i + 4] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[24].x, UpLeft[24].y, DotState[i + 4] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[25].x, UpLeft[25].y, DotState[i + 5] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[26].x, UpLeft[26].y, DotState[i + 5] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[27].x, UpLeft[27].y, DotState[i + 5] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[28].x, UpLeft[28].y, DotState[i + 5] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[29].x, UpLeft[29].y, DotState[i + 5] and $01 <> 0);
          DrawDot(WhereTo, UpLeft[30].x, UpLeft[30].y, DotState[i + 6] and $10 <> 0);
          DrawDot(WhereTo, UpLeft[31].x, UpLeft[31].y, DotState[i + 6] and $08 <> 0);
          DrawDot(WhereTo, UpLeft[32].x, UpLeft[32].y, DotState[i + 6] and $04 <> 0);
          DrawDot(WhereTo, UpLeft[33].x, UpLeft[33].y, DotState[i + 6] and $02 <> 0);
          DrawDot(WhereTo, UpLeft[34].x, UpLeft[34].y, DotState[i + 6] and $01 <> 0);
        end;
      end;
    ds5x7Bulb:
      begin
        with WhereTo.Canvas do
        begin
          i := ord(FDigit) * 7;
          DrawBulb(WhereTo, UpLeft[0].x, UpLeft[0].y, DotState[i] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[1].x, UpLeft[1].y, DotState[i] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[2].x, UpLeft[2].y, DotState[i] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[3].x, UpLeft[3].y, DotState[i] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[4].x, UpLeft[4].y, DotState[i] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[5].x, UpLeft[5].y, DotState[i + 1] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[6].x, UpLeft[6].y, DotState[i + 1] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[7].x, UpLeft[7].y, DotState[i + 1] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[8].x, UpLeft[8].y, DotState[i + 1] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[9].x, UpLeft[9].y, DotState[i + 1] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[10].x, UpLeft[10].y, DotState[i + 2] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[11].x, UpLeft[11].y, DotState[i + 2] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[12].x, UpLeft[12].y, DotState[i + 2] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[13].x, UpLeft[13].y, DotState[i + 2] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[14].x, UpLeft[14].y, DotState[i + 2] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[15].x, UpLeft[15].y, DotState[i + 3] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[16].x, UpLeft[16].y, DotState[i + 3] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[17].x, UpLeft[17].y, DotState[i + 3] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[18].x, UpLeft[18].y, DotState[i + 3] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[19].x, UpLeft[19].y, DotState[i + 3] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[20].x, UpLeft[20].y, DotState[i + 4] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[21].x, UpLeft[21].y, DotState[i + 4] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[22].x, UpLeft[22].y, DotState[i + 4] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[23].x, UpLeft[23].y, DotState[i + 4] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[24].x, UpLeft[24].y, DotState[i + 4] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[25].x, UpLeft[25].y, DotState[i + 5] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[26].x, UpLeft[26].y, DotState[i + 5] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[27].x, UpLeft[27].y, DotState[i + 5] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[28].x, UpLeft[28].y, DotState[i + 5] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[29].x, UpLeft[29].y, DotState[i + 5] and $01 <> 0);
          DrawBulb(WhereTo, UpLeft[30].x, UpLeft[30].y, DotState[i + 6] and $10 <> 0);
          DrawBulb(WhereTo, UpLeft[31].x, UpLeft[31].y, DotState[i + 6] and $08 <> 0);
          DrawBulb(WhereTo, UpLeft[32].x, UpLeft[32].y, DotState[i + 6] and $04 <> 0);
          DrawBulb(WhereTo, UpLeft[33].x, UpLeft[33].y, DotState[i + 6] and $02 <> 0);
          DrawBulb(WhereTo, UpLeft[34].x, UpLeft[34].y, DotState[i + 6] and $01 <> 0);
        end;
      end;
    ds16Seg:
      begin
        with WhereTo.Canvas do
        begin
          i := ord(FDigit);
          DrawTopSegment(WhereTo, UpLeft[0].x, UpLeft[0].y, SegHorLen, FSegWidth,
                         SixState[i] and $0001 <> 0);
          DrawTopSegment(WhereTo, UpLeft[1].x, UpLeft[1].y, SegHorLen, FSegWidth,
                         SixState[i] and $0002 <> 0);
          DrawLeftSegment(WhereTo, UpLeft[2].x, UpLeft[2].y, FSegWidth, SegVerLen,
                          SixState[i] and $0004 <> 0);
          DrawNorthWest(WhereTo, UpLeft[3].x, UpLeft[3].y, UpLeft[23].x, UpLeft[23].y,
                        SixState[i] and $0008 <> 0);
          Hexagon(WhereTo, UpLeft[4].x, UpLeft[4].y, FSegWidth, SegVerLen,
                  SixState[i] and $0010 <> 0);
          DrawNorthEast(WhereTo, UpLeft[5].x, UpLeft[5].y, UpLeft[25].x, UpLeft[25].y,
                        SixState[i] and $0020 <> 0);
          DrawRightSegment(WhereTo, UpLeft[6].x, UpLeft[6].y, FSegWidth, SegVerLen,
                           SixState[i] and $0040 <> 0);
          Hexagon(WhereTo, UpLeft[7].x, UpLeft[7].y, SegHorLen, FSegWidth,
                  SixState[i] and $0080 <> 0);
          Hexagon(WhereTo, UpLeft[8].x, UpLeft[8].y, SegHorLen, FSegWidth,
                  SixState[i] and $0100 <> 0);
          DrawLeftSegment(WhereTo, UpLeft[9].x, UpLeft[9].y, FSegWidth, SegVerLen,
                          SixState[i] and $0200 <> 0);
          DrawNorthEast(WhereTo, UpLeft[10].x, UpLeft[10].y, UpLeft[30].x, UpLeft[30].y,
                        SixState[i] and $0400 <> 0);
          Hexagon(WhereTo, UpLeft[11].x, UpLeft[11].y, FSegWidth, SegVerLen,
                  SixState[i] and $0800 <> 0);
          DrawNorthWest(WhereTo, UpLeft[12].x, UpLeft[12].y, UpLeft[32].x, UpLeft[32].y,
                        SixState[i] and $1000 <> 0);
          DrawRightSegment(WhereTo, UpLeft[13].x, UpLeft[13].y, FSegWidth, SegVerLen,
                           SixState[i] and $2000 <> 0);
          DrawBottomSegment(WhereTo, UpLeft[14].x, UpLeft[14].y, SegHorLen, FSegWidth,
                            SixState[i] and $4000 <> 0);
          DrawBottomSegment(WhereTo, UpLeft[15].x, UpLeft[15].y, SegHorLen, FSegWidth,
                            SixState[i] and $8000 <> 0);
        end;
      end;
  end;
end;

procedure TDisplay.Paint;
var
  R: TRect;
  PaintBMP: TBitmap;
begin
  inherited Paint;
  if (OldHeight <> Height) or (OldWidth <> Width) then
  begin
    SetMeasures;
    OldHeight := Height;
    OldWidth := Width;
  end;
  PaintBMP := TBitmap.Create;
  PaintBMP.Height := Height - 2;
  PaintBMP.Width := Width - 2;
  with PaintBMP.Canvas do
  begin
    Brush.Color := FBackColor;
    R := ClientRect;
    dec(R.Right, 2);
    dec(R.Bottom, 2);
    FillRect(R);
    DrawDigit(PaintBMP);
  end;
  Canvas.CopyMode := cmSrcCopy;
  Canvas.Draw(1, 1, PaintBMP);
  PaintBMP.Free;
end;

procedure Register;
begin
  RegisterComponents('FreeWare', [TDisplay]);
end;

end.
