unit FileSrc;

{ TFileSource component by Peter Martinsson (1997.01.01)

  FREEWARE

  The purpose of this component is to be able to store ANY kind of file
  in the .EXE file. This is useful if a component needs a file and
  you don't want to supply the extra file.

  Properties:
    Compress: Boolean    - set to True if you want to compress the stored data
                           (reduces EXE-size)
    DisplayLabel: string - an optional string
    Extension: string    - extension of the created tempfile with '.' (.TMP)
    FileName: TFileName  - the file
    Store: Boolean       - if True, the file in FileName will be stored in the
                           .EXE file and at run-time the FileName property
                           will point to a temporary file (c:\WIN95\TEMP\~fs???+Extension),
                           which will be removed on destruction
                           if False, FileName will point to the original file and
                           you will have to distribute the file as normal

    The compression algoritm used is described in the LZRW1KH unit. It is not
    the best algoritm but it is fast. I have wrapped it up in the
    TCompressStream object (see the source below).


  -- Use this unit on your own risk. --

  If you have any comments please send me an e-mail:
    martinsson@mbox301.swipnet.se
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DsgnIntf;

type
  TFileSource = class(TComponent)
  private
    { Private declarations }
    FCompress: Boolean;
    FDisplayLabel: string;
    FExtension: string;
    FFileName: TFileName;
    FStore: Boolean;
    FTempFileCreated: Boolean;
    function CreateTempFileName: string;
    function GetFileName: TFileName;
    function StoreFileName: Boolean;
    procedure DeleteTempFile;
    procedure SetCompress(Value: Boolean);
    procedure SetFileName(Value: TFileName);
  protected
    { Protected declarations }
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadFileData(Stream: TStream); virtual;
    procedure WriteFileData(Stream: TStream); virtual;
  public
    { Public declarations }
    destructor Destroy; override;
  published
    { Published declarations }
    property Compress: Boolean read FCompress write SetCompress;
    property DisplayLabel: string read FDisplayLabel write FDisplayLabel;
    property Extension: string read FExtension write FExtension;
    property FileName: TFileName read GetFileName write SetFileName stored StoreFileName;
    property Store: Boolean read FStore write FStore;
  end;

  TCompressStream = class(TObject)
  private
    FSignature: array[0..5] of Char; {= 'PMCS10' }
  public
    constructor Create; virtual;
    procedure Compress(InStream, OutStream: TStream; InSize: LongInt);
    procedure DeCompress(InStream, OutStream: TStream);
  end;

  TFileNameProperty = class(TPropertyEditor)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

procedure Register;

implementation

uses
  LZRW1KH;

procedure Register;
begin
  RegisterComponents('Samples', [TFileSource]);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileSource, 'FileName', TFileNameProperty);
end;

{ TCompressStream }

constructor TCompressStream.Create;
begin
  FSignature := 'PMCS10';
end;

procedure TCompressStream.Compress(InStream, OutStream: TStream; InSize: LongInt);
var InBuffer, OutBuffer: BufferArray;
    CompressedSize, BytesRead, FinalPos, SizePos, TotalSize: LongInt;
begin
  TotalSize := 0;
  OutStream.WriteBuffer(FSignature, SizeOf(FSignature));
  SizePos := OutStream.Position;
  OutStream.WriteBuffer(TotalSize, SizeOf(TotalSize));
  while InSize > 0 do
  begin
    BytesRead := InStream.Read(InBuffer, SizeOf(InBuffer));
    CompressedSize := Compression(@InBuffer, @OutBuffer, BytesRead);
    OutStream.WriteBuffer(CompressedSize, SizeOf(CompressedSize));
    OutStream.WriteBuffer(OutBuffer, CompressedSize);
    TotalSize := TotalSize + CompressedSize + SizeOf(CompressedSize);
    InSize := InSize - BytesRead;
  end;
  FinalPos := OutStream.Position;
  OutStream.Position := SizePos;
  OutStream.WriteBuffer(TotalSize, SizeOf(TotalSize));
  OutStream.Position := FinalPos;
end;

procedure TCompressStream.DeCompress(InStream, OutStream: TStream);
var InBuffer, OutBuffer: BufferArray;
    CompressedSize, UnCompressedSize, InSize: LongInt;
    Sig: array[0..SizeOf(FSignature)-1] of Char;
begin
  InStream.ReadBuffer(Sig, SizeOf(FSignature));
  if Sig <> FSignature then raise Exception.Create('Wrong file type');
  InStream.ReadBuffer(InSize, SizeOf(InSize));
  while InSize > 0 do
  begin
    InStream.ReadBuffer(CompressedSize, SizeOf(CompressedSize));
    InStream.ReadBuffer(InBuffer, CompressedSize);
    UnCompressedSize := DeCompression(@InBuffer, @OutBuffer, CompressedSize);
    OutStream.WriteBuffer(OutBuffer, UnCompressedSize);
    InSize := InSize - CompressedSize - SizeOf(CompressedSize);
  end;
end;

{ TFileSource }

destructor TFileSource.Destroy;
begin
  DeleteTempFile;
  inherited Destroy;
end;

procedure TFileSource.DeleteTempFile;
begin
  if FTempFileCreated then
    DeleteFile(FFileName);
end;

procedure TFileSource.SetCompress(Value: Boolean);
begin
  FCompress := Value;
  if Value then
    Store := True;
end;

function TFileSource.GetFileName: TFileName;
begin
  if FileExists(FFileName) then
    Result := FFileName
  else
    Result := '';
end;

procedure TFileSource.SetFileName(Value: TFileName);
begin
  DeleteTempFile;
  FFileName := Value;
  FTempFileCreated := False;
  FExtension := ExtractFileExt(Value);
end;

function TFileSource.StoreFileName: Boolean;
begin
  Result := not FStore;
end;

function TFileSource.CreateTempFileName: string;
var TempDir: array[1..MAX_PATH] of Char;
    count, len: LongInt;
    dummy, sec: Word;
    start: string;
begin
  count := 0;
  DecodeTime(Now, dummy, dummy, sec, dummy);
  len := GetTempPath(MAX_PATH, @TempDir);
  if TempDir[len] = '\' then TempDir[len] := #0;
  start := StrPas(@TempDir) + '\~fs' + IntToStr(sec);
  while FileExists(start + IntToStr(count) + FExtension) do
    Inc(count);
  Result := start + IntToStr(count) + FExtension;
  FileClose(FileCreate(Result));
end;

procedure TFileSource.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('FileData', ReadFileData, WriteFileData, FStore);
end;

procedure TFileSource.ReadFileData(Stream: TStream);
var Size: LongInt;
    FileStream: TFileStream;
begin
  FFileName := CreateTempFileName;
  FTempFileCreated := True;
  FileStream := TFileStream.Create(FFileName, fmOpenWrite);
  try
    if FCompress then
      with TCompressStream.Create do
      try
        DeCompress(Stream, FileStream);
      finally
        Free;
      end
    else
    begin
      Stream.ReadBuffer(Size, SizeOf(Size));
      FileStream.CopyFrom(Stream, Size);
    end;
  finally
    FileStream.Free;
  end;
end;

procedure TFileSource.WriteFileData(Stream: TStream);
var Size: LongInt;
    FileStream: TFileStream;
begin
  if FileExists(FFileName) then
  begin
    FileStream := TFileStream.Create(FFileName, fmOpenRead);
    try
      Size := FileStream.Size;
      if FCompress then
      begin
        with TCompressStream.Create do
        try
          Compress(FileStream, Stream, Size);
        finally
          Free;
        end;
      end else
      begin
        Stream.WriteBuffer(Size, SizeOf(Size));
        Stream.CopyFrom(FileStream, Size);
      end;
    finally
      FileStream.Free;
    end;
  end else
    raise Exception.Create(Name + ' cannot store file: ' + FFileName + ' (file is missing)');
end;

{ Property editor. }

procedure TFileNameProperty.Edit;
begin
  with TOpenDialog.Create(nil) do
  try
    FileName := GetStrValue;
    if Execute then
      SetStrValue(FileName);
  finally
    Free;
  end;
end;

function TFileNameProperty.GetValue: string;
begin
  Result := GetStrValue;
end;

procedure TFileNameProperty.SetValue(const Value: string);
begin
  SetStrValue(Value);
end;

function TFileNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

end.
