unit Matrices;

interface

uses SysUtils, DsgnIntf;

type
EMatrixError=class(Exception);

TContents=array[0..5000] of double;
PContents=^TContents;

TMatrix=class(TObject)
private
   FContents:PContents;
   FRows:integer;
   FColumns:integer;
   FTransposed:boolean;
   function FGetContents(GRow,GColumn:integer):double;
   procedure FSetContents(GRow,GColumn:integer; Value:double);
   function FGetRows:integer;
   function FGetColumns:integer;
public
   constructor Create(CRows,CColumns:integer);
   constructor Multiply(const A,B:TMatrix);
   constructor CreateIdentity(Size:integer);
   constructor RotX(phi:double);
   constructor RotY(phi:double);
   constructor RotZ(phi:double);
   procedure Clear;
   procedure Add(const M:TMatrix);
   procedure Sub(const M:TMatrix);
   procedure Assign(const M:TMatrix);
   procedure DivScal(Fact:double);
   procedure MulScal(Fact:double);
   function Norm:double;
   procedure Free;
   procedure Transpose;
   property Contents[Index:integer;Index:integer]:double read FGetContents
            write FSetContents; default;
   property Rows:integer read FGetRows;
   property Columns:integer read FGetColumns;
end;

TVector=class(TMatrix)
private
   function FGetContents(GIndex:integer):double;
   procedure FSetContents(GIndex:integer; Value:double);
public
   constructor Create(CSize:integer);
   constructor CreateWithCoords(Coords:array of double);
   property Contents[Index:integer]:double read FGetContents
            write FSetContents;
end;

TMatrixProperty=class(TClassProperty)
   procedure Edit; override;
   function GetAttributes: TPropertyAttributes; override;
end;

procedure Register;

implementation
uses MatrixEd, Forms;

constructor TMatrix.Create(CRows,CColumns:integer);
begin
inherited Create;
FRows:=CRows;
FColumns:=CColumns;
GetMem(FContents,FRows*FColumns*sizeof(double));
FTransposed:=False;
Clear;
end;

constructor TMatrix.CreateIdentity(Size:integer);
var i:integer;
begin
Create(Size,Size);
Clear;
for i:=0 to Size-1 do
    Contents[i,i]:=1;
end;

constructor TMatrix.RotX(phi:double);
var s,c:double;
begin
Create(3,3);
c:=cos(phi);
s:=sin(phi);
Contents[0,0]:=1 ;Contents[0,1]:=0 ;Contents[0,2]:=0 ;
Contents[1,0]:=0 ;Contents[1,1]:=c ;Contents[1,2]:=-s ;
Contents[2,0]:=0 ;Contents[2,1]:=s ;Contents[2,2]:=c ;
end;

constructor TMatrix.RotY(phi:double);
var s,c:double;
begin
Create(3,3);
c:=cos(phi);
s:=sin(phi);
Contents[0,0]:=c  ;Contents[0,1]:=0 ;Contents[0,2]:=s ;
Contents[1,0]:=0  ;Contents[1,1]:=1 ;Contents[1,2]:=0 ;
Contents[2,0]:=-s ;Contents[2,1]:=0 ;Contents[2,2]:=c ;
end;

constructor TMatrix.RotZ(phi:double);
var s,c:double;
begin
Create(3,3);
c:=cos(phi);
s:=sin(phi);
Contents[0,0]:=c ;Contents[0,1]:=-s ;Contents[0,2]:=0 ;
Contents[1,0]:=s ;Contents[1,1]:=c  ;Contents[1,2]:=0 ;
Contents[2,0]:=0 ;Contents[2,1]:=0  ;Contents[2,2]:=1 ;
end;

procedure TMatrix.Free;
begin
FreeMem(FContents,FRows*FColumns*sizeof(double));
inherited Free;
end;

procedure TMatrix.Clear;
var i,j:integer;
begin
for i:=0 to Rows-1 do
    for j:=0 to Columns-1 do
        Contents[i,j]:=0;
end;

function TMatrix.FGetContents(GRow,GColumn:integer):double;
begin
if (GRow<0) or (GRow>=Rows) or (GColumn<0) or (GColumn>=Columns) then
   raise EMatrixError.Create('Index out of range')
else begin
   if not FTransposed then
      Result:=FContents^[GRow*FColumns+GColumn]
   else
   Result:=FContents^[GColumn*FColumns+GRow];
   end
end;

procedure TMatrix.FSetContents(GRow,GColumn:integer; Value:double);
begin
if (GRow<0) or (GRow>=Rows) or (GColumn<0) or (GColumn>=Columns) then
   raise EMatrixError.Create('Index out of range')
else begin
   if not FTransposed then
      FContents^[GRow*FColumns+GColumn]:=Value
   else
      FContents^[GColumn*FColumns+GRow]:=Value;
   end;
end;

function TMatrix.FGetRows:integer;
begin
if not FTransposed then
   Result:=FRows
else
   Result:=FColumns;
end;

function TMatrix.FGetColumns:integer;
begin
if not FTransposed then
   Result:=FColumns
else
   Result:=FRows;
end;

constructor TMatrix.Multiply(const A,B:TMatrix);
var i,j,k:integer;
begin
if A.Columns<>B.Rows then
   raise EMatrixError.Create('Matrix size do not match in multiply')
else begin
   Create(A.Rows,B.Columns);
   for i:=0 to A.Rows-1 do begin
       for j:=0 to B.Columns-1 do begin
           Contents[i,j]:=0;
           for k:=0 to A.Columns-1 do begin
               Contents[i,j]:=Contents[i,j]+A[i,k]*B[k,j];
               end;
           end;
       end;
   end;
end;

procedure TMatrix.Add(const M:TMatrix);
var i,j:integer;
begin
if (M.Rows<>Rows) or (M.Columns<>Columns) then
   raise EMatrixError.Create('Matrix size does not match in Add')
else begin
   for i:=0 to Rows-1 do
       for j:=0 to Columns-1 do
           Contents[i,j]:=Contents[i,j]+M[i,j];
   end;
end;

procedure TMatrix.Sub(const M:TMatrix);
var i,j:integer;
begin
if (M.Rows<Rows) or (M.Columns<Columns) then
   raise EMatrixError.Create('Matrix size does not match in Sub')
else begin
   for i:=0 to Rows-1 do
       for j:=0 to Columns-1 do
           Contents[i,j]:=Contents[i,j]-M[i,j];
   end;
end;

procedure TMatrix.Assign(const M:TMatrix);
var i,j:integer;
begin
if (M.Rows<Rows) or (M.Columns<Columns) then
   raise EMatrixError.Create('Matrix size does not match in Assign')
else begin
   for i:=0 to Rows-1 do
       for j:=0 to Columns-1 do
           Contents[i,j]:=M[i,j];
   end;
end;

procedure TMatrix.DivScal(Fact:double);
var i,j:integer;
begin
for i:=0 to Rows-1 do
    for j:=0 to Columns-1 do
        Contents[i,j]:=Contents[i,j]/Fact;
end;

procedure TMatrix.MulScal(Fact:double);
var i,j:integer;
begin
for i:=0 to Rows-1 do
    for j:=0 to Columns-1 do
        Contents[i,j]:=Contents[i,j]*Fact;
end;

function TMatrix.Norm:double;
var i,j:integer;
begin
Result:=0;
for i:=0 to Rows-1 do
    for j:=0 to Columns-1 do
        Result:=Result+sqr(Contents[i,j]);
Result:=sqrt(Result);
end;

procedure TMatrix.Transpose;
begin
FTransposed:=not FTransposed;
end;

constructor TVector.Create(CSize:integer);
begin
inherited Create(CSize,1);
end;

constructor TVector.CreateWithCoords(Coords:array of double);
var i:integer;
begin
if High(Coords)>=0 then begin
   Create(High(Coords)+1);
   for i:=0 to High(Coords) do Contents[i]:=Coords[i];
   end
else raise EMatrixError.Create('Invalid array in TVector.CreateWithCoords');
end;

function TVector.FGetContents(GIndex:integer):double;
begin
Result:= inherited FGetContents(GIndex,0);
end;

procedure TVector.FSetContents(GIndex:integer; Value:double);
begin
inherited FSetContents(GIndex,0,Value);
end;

procedure TMatrixProperty.Edit;
var Form:TMatrixForm;
    M:TMatrix;
begin
M:=TMatrix(GetOrdValue);
Form:=TMatrixForm.Create(Application);
try
  Form.ShowModal(M);
finally
  Form.Free;
end;
end;

function TMatrixProperty.GetAttributes: TPropertyAttributes;
begin
Result:=[paDialog]
end;

procedure Register;
begin
RegisterPropertyEditor(Typeinfo(TMatrix),nil,'',TMatrixProperty);
end;

end.
