
{*****************************************************************************}
{                                                                             }
{            QDB v1.04 Visual Components for Delphi 1.x & 2.x                 }
{                                                                             }
{          Copyright (c) 1995, 1996, 1997 Robert R. Marsh, S.J.               }
{             & the British Province of the Society of Jesus                  }
{                                                                             }
{        Portions of this code are based on the Delphi VCL library            }
{                                                                             }
{              This source code may *not* be redistributed                    }
{              ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                    }
{                                                                             }
{       If you like QDB and find yourself using it please consider            }
{       making a donation to your favorite charity. I would also be           }
{       pleased if you would acknowledge QDB in any projects that             }
{       make use of it.                                                       }
{                                                                             }
{       QDB is supplied as is. The author disclaims all warranties,           }
{       expressed or implied, including, without limitation, the              }
{       warranties of merchantability and of fitness for any purpose.         }
{       The author assumes no liability for damages, direct or                }
{       consequential, which may result from the use of QDB.                  }
{                                                                             }
{*****************************************************************************}

(*
  Watch out for compiler warnings after try ... except blocks. The compiler
  doesn't know that the File/IndexError functions raise exceptions of their
  own which prevents the following code from ever being executed uninitialized.
*)

(*

  HISTORY of Modifications

  Version 1.04
    Corrected an exception handling problem revealed in the 16-bit version
    Fixed bug in filter for repeated patterns
    Simplified the buffered file handling of the Index
    Made it less likely for files to be overwritten incorrectly
    Rationalized handling of ReadOnly files
    Tidied up monitoring the progress of processes
    Fixed bug with sorting index (16-bit only)
    Prevented operations when FileName = '' & added Ready property

  Version 1.03 ~
    Source modified to compile under both Delphi 1 & 2

  Version 1.02 ~
    No changes to the code -- recompiled with new SYSTEM.DCU

  Version 1.01 ~
    Fixed bug in SaveAs -- didn't update FFileAge
    Fixed bug in Kill -- didn't kill!
    Changed timing function of the cache for speed's sake
    Fixed bug in Matches -- failed if pattern occurs more than once
    Fixed bug with EoF and BoF -- failed if next key matches
    Added two methods to TMyStringList and quartered the load time!
    Modified Save to match SaveIndex
    Added FileToRecover, OrphanToRecover & Recover methods ...

*)

unit QDB;

interface

uses
  SysUtils, Classes, Messages, Controls, Forms,
  ExtCtrls, Buttons, DsgnIntf,
  {$IFNDEF WIN32}WinTypes, WinProcs;{$ELSE}Windows;{$ENDIF}

type
  string40 = string [40];
  string05 = string [5];

const
  FileVersion: String05 = '1.04';
  AuthorInfo : String40 = 'Robert R. Marsh, SJ -- RobMarsh@aol.com';


{ basic types that your application needs to know about }

type
  TKey         = string[255];    { QDB key                   }
  TItemIndex   = integer;        { pointer into index        }
  TDataIndex   = longint;        { pointer into item of data }
  TQDBFileName = string;
  TFileHandle  = integer;
  TFilePos     = longint;        { pointer into disk file    }


{ event used to signal progress of lengthy process }

type
  TPercentage     = 0..100;
  TProgressOrigin = (prStart,prFinish,prSave,prPack,prKeyList);
  TProgressEvent  = procedure(Sender : TObject;
                              Percent: TPercentage;
                              Kind   : TProgressOrigin) of object;


{ event used to get confirmation from user }

type
  TConfirmEvent   = procedure(Sender: TObject; var OK: boolean) of object;


{ event used to give warning to the user }

type
  TWarningEvent   = TNotifyEvent;


{ QDB-specific exceptions }

type
  EQDBError      = class(Exception);
  EQDBFileError  = class(EQDBError);
  EQDBIndexError = class(EQDBError);
  EQDBListError  = class(EQDBError);

{$IFDEF WIN32}
{ MyStringList is a slightly modified StringList which publishes its  }
{ SetCapacity method. Code from the VCL library is Copyright Borland. }
{ Copyright (c) 1995, 1996 Borland International. All Rights Reserved.}

type
  PStringItem = ^TStringItem;
  TStringItem = record
    FString: string;
    FObject: TObject;
  end;

type
  PStringItemList = ^TStringItemList;
  TStringItemList = array[0..MaxListSize] of TStringItem;


  TMyStringList = class(TStrings)
  private
    FList: PStringItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FDuplicates: TDuplicates;
    FUpdateCount: Integer;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer);
    procedure InsertItem(Index: Integer; const S: string);
    procedure SetSorted(Value: Boolean);
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetUpdateState(Updating: Boolean); override;
    procedure Reverse;     { reverses the items of an unsorted list }
    procedure PretendSort; { sets sorted to true without sorting    }
  public
    destructor Destroy; override;
    { the capacity functions have been moved from the private block }
    function GetCapacity: integer;
    procedure SetCapacity(NewCapacity: TItemIndex);
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: string; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: string): Integer; override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Sort; virtual;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted write SetSorted;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;
{$ELSE}
  TMyStringList = class(TStringList)
  public
    function GetCapacity: integer;
    procedure SetCapacity(NewCapacity: integer);
  end;
{$ENDIF}


{ these two list types each know how to destroy their objects }

type
  TIndexList = class(TMyStringList)
    destructor Destroy; override;
    procedure EmptyAndClear;
  end;

  TCacheList = class(TStringList)
    destructor Destroy; override;
    procedure EmptyAndClear;
  end;

{ buffered file stream }

type
  TBFStream = class(TFileStream)
  private
    buffer      : pchar;
    bufferpos   : pchar;
    buflen      : longint;
    bytesread   : longint;
    isdirty     : Boolean;
  public
    {if BufferSize = -1 a buffer as big as the file will be used}
    constructor Create(const FileName : string; Mode : Word; BufferSize: longint);
    destructor Destroy; override;
    function Seek(offset: longint; origin: word ): longint; override;
    procedure ResetBuffer;
    procedure AdjustBuffer;
    function GetLongint(var L: longint): boolean;
    procedure PutLongint(const L: longint);
    function GetKey(var K: TKey): boolean;
    procedure PutKey(const K: TKey);
  end;

type
  TTempBFStream = class(TBFStream)
  private
    FOldFileName: string;
    TmpFileName : string;
  public
    constructor Create(const OldFileName: string);
    destructor Destroy; override;
  end;


{ the QDB component }

type
  TQDB = class(TComponent)
  private
    FAuthor          : string40;    { my name!                                      }
    FVersion         : string05;    { QDB version e.g. '1.00'                       }
    FReady           : boolean;     { true iff a file is open and ready for access  }
    QDBFile          : TFileHandle; { working data-file handle                      }
    QIXFile          : TBFStream;   { working index-file stream                     }
    Index            : TIndexList;  { in-memory index file                          }
    Cache            : TCacheList;  { in-memory item cache                          }
    CurrentCacheSize : longint;     { ... in bytes                                  }
    FCacheHits       : longint;     { number of cached accesses                     }
    FCacheTotal      : longint;     { total number of accesses                      }
    FCacheSize       : longint;     { upper limit on cache                          }
    IsDirty          : boolean;     { has the file been changed ?                   }
    FItemIndex       : TItemIndex;  { current position in Index                     }
    FFilename        : string;      { name of QDB file                              }
    FFileAge         : longint;     { age of the QDB file when opened               }
    QDBFileName      : string;      { name of the working data-file                 }
    QIXFileName      : string;      { name of the working index-file                }
    FCount           : TItemIndex;  { number of items in Index                      }
    FKey             : TKey;        { key of current item                           }
    FReadOnly        : boolean;     { governs file access                           }
    FBoF             : boolean;     { at beginning of file ?                        }
    FEoF             : boolean;     { at end of file ?                              }
    FItemSize        : TDataIndex;  { of the current item                           }
    FMaxItemSize     : TDataIndex;  { of the largest item                           }
    FFilter          : TKey;        { restricts access to certain keys              }
    Filtered         : boolean;     { is FFilter = '' ?                             }
    FFrontWild       : char;        { wildcard stands for any chars at front of key }
    FBackWild        : char;        { wildcard stands for any chars at back of key  }
    FUpdating        : integer;     { count Begin/End Update calls                  }
    MonitorInterval  : longint;     { how often to update progress monitor          }
    MonitorKind      : TProgressOrigin;
    FOnNavigate      : TNotifyEvent;
    FOnAdded         : TNotifyEvent;
    FOnDeleted       : TNotifyEvent;
    FOnChanged       : TNotifyEvent;
    FOnFound         : TNotifyEvent;
    FOnFileAssigned  : TNotifyEvent;
    FWarnOutOfBounds : TWarningEvent;
    FWarnNoData      : TWarningEvent;
    FWarnReadOnly    : TWarningEvent;
    FWarnCannotCache : TWarningEvent;
    FProgressUpdate  : TProgressEvent;
    FBeforeKill      : TConfirmEvent;
    FBeforeOverWrite : TConfirmEvent;
    FOnKilled        : TNotifyEvent;
    function  GetKey(value: TItemIndex): TKey;
    function  GetItemSize(value: TItemIndex): TDataIndex;
    function  GetFileName: TQDBFileName;
    procedure SetFileName(value: TQDBFileName);
    procedure SetItemIndex(value: TItemIndex);
    procedure SetFilter(value: TKey);
    procedure SetCacheSize(value: longint);
    function  GetCacheFrequency: integer;
    procedure SetCacheFrequency(value: integer);
    procedure SetDummyAuthor(value: string40);
    procedure SetDummyVersion(value: string05);
    procedure FileError(ErrCode: integer; SDefault: integer);
    procedure IndexError(SErr: integer);
    procedure Split;
    procedure Splice;
    procedure CreateQDB;
    procedure OpenQDB;
    procedure CloseQDB;
    procedure LoadIndex;
    procedure SaveIndex;
    procedure EnCache(ItemPtr: pointer; ItemLen: TDataIndex);
    procedure DeCache;
    procedure UnCache;
    procedure FlushCache;
    procedure MonitorSetup(const max: longint; const kind: TProgressOrigin);
    procedure MonitorZero;
    procedure MonitorMax;
    procedure MonitorUpdate(const n: longint);
  protected
    procedure Navigate;
    procedure Added;
    procedure Deleted;
    procedure Changed;
    procedure Found;
    procedure FileAssigned;
    procedure OutOfBounds;
    procedure NoData;
    procedure CannotChange;
    procedure CannotCache;
    procedure SignalProgress(Percent: TPercentage; Kind: TProgressOrigin);
    procedure AboutToKill(var OK: boolean);
    procedure AboutToOverWrite(var OK: boolean);
    procedure Killed;
    function  Matches(key: TKey; pattern: TKey): boolean;
    function  FileToRecover: string;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure AssignKeyList(Keys: TStrings);
    procedure FirstItem;
    procedure LastItem;
    procedure NextItem;
    procedure PrevItem;
    procedure GetItem(ItemPtr: pointer);
    function  ExactMatch(key: TKey): boolean;
    function  CloseMatch(partialkey: TKey): boolean;
    function  PatternMatch(pattern: TKey): boolean;
    procedure SetMatchChars(front: char; back: char);
    procedure PrepareToAdd(numberofitems: integer);
    procedure AddItem(ItemPtr: pointer; ItemLen: TDataIndex; key: TKey);
    procedure DeleteItem;
    procedure ChangeItem(ItemPtr: pointer; ItemLen: TDataIndex);
    procedure AddStreamItem(stream: TStream; key: TKey);
    procedure GetStreamItem(stream: TStream);
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Pack;
    procedure Save;
    procedure SaveAs(NewName: string);
    procedure Kill;
    function  OrphanToRecover: boolean;
    procedure Recover(NewFileName: string);
    { read-only run-time properties }
    property Ready:         boolean read FReady;
    property ItemIndex:     TItemIndex read FItemIndex write SetItemIndex;
    property ItemSize:      TDataIndex read FItemSize;
    property Count:         TItemIndex read FCount;
    property Key:           TKey read FKey;
    property MaxItemSize:   TDataIndex read FMaxItemSize;
    property BoF:           boolean read FBoF;
    property EoF:           boolean read FEoF;
  published
    property FileName:         TQDBFileName read GetFileName write SetFileName;
    property ReadOnly:         boolean read FReadOnly write FReadOnly;
    property Filter:           TKey read FFilter write SetFilter;
    property CacheSize:        longint read FCacheSize write SetCacheSize;
    property CacheFrequency:   integer read GetCacheFrequency write SetCacheFrequency;
    property OnNavigate:       TNotifyEvent read FOnNavigate write FOnNavigate;
    property OnAdded:          TNotifyEvent read FOnAdded write FOnAdded;
    property OnDeleted:        TNotifyEvent read FOnDeleted write FOnDeleted;
    property OnChanged:        TNotifyEvent read FOnChanged write FOnChanged;
    property OnFound:          TNotifyEvent read FOnFound write FOnFound;
    property OnFileAssigned:   TNotifyEvent read FOnFileAssigned write FOnFileAssigned;
    property WarnOutOfBounds:  TWarningEvent read FWarnOutOfBounds write FWarnOutOfBounds;
    property WarnNoData:       TWarningEvent read FWarnNoData write FWarnNoData;
    property WarnReadOnly:     TWarningEvent read FWarnReadOnly write FWarnReadOnly;
    property WarnCannotCache:  TWarningEvent read FWarnCannotCache write FWarnCannotCache;
    property ProgressUpdate:   TProgressEvent read FProgressUpdate write FProgressUpdate;
    property BeforeKill:       TConfirmEvent read FBeforeKill write FBeforeKill;
    property BeforeOverWrite:  TConfirmEvent read FBeforeOverWrite write FBeforeOverWrite;
    property OnKilled:         TNotifyEvent read FOnKIlled write FOnKilled;
    { the following two properties should really be public }
    { rather than published but I want them to show up in  }
    { the Object Inspector so I put them here with a dummy }
    { routine to set the property.                         }
    property AboutVersion: string05 read FVersion write SetDummyVersion;
    property AboutAuthor:  string40 read FAuthor write SetDummyAuthor;
  end;


{ TQDBFilenameProperty -- Property editor for the TQDBFileName property. }
{ Displays a File Open Dialog for the name of the QDB file.              }

type
  TQDBFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;


{ TQDBNavigator }

{ This is a modified version of DBNavigator. Code from the VCL library }
{ is copyright Borland.                                                }
{ Copyright (c) 1995, 1996 Borland International. All Rights Reserved. }

type
  TNavButton = class;

  TNavGlyph       = (ngEnabled, ngDisabled);
  TNavigateBtn    = (nbFirst, nbPrev, nbNext, nbLast);
  TButtonSet      = set of TNavigateBtn;
  TNavButtonStyle = set of (nsAllowTimer, nsFocusRect);

  TNavClickEvent  = procedure (Sender: TObject; Button: TNavigateBtn) of object;

  TQDBNavigator = class (TCustomPanel)
  private
    FQDB: TQDB;
    FVisibleButtons: TButtonSet;
    FHints: TStrings;
    ButtonWidth: Integer;
    MinBtnSize: TPoint;
    FOnNavClick: TNavClickEvent;
    FocusedButton: TNavigateBtn;
    procedure InitButtons;
    procedure InitHints;
    procedure _Click(Sender: TObject);
    procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetVisible(Value: TButtonSet);
    procedure AdjustSize (var W: Integer; var H: Integer);
    procedure SetHints(Value: TStrings);
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    Buttons: array[TNavigateBtn] of TNavButton;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure BtnClick(Index: TNavigateBtn);
  published
    property VisibleButtons: TButtonSet read FVisibleButtons write SetVisible
      default [nbFirst, nbPrev, nbNext, nbLast];
    property QDB: TQDB read FQDB write FQDB;
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Ctl3D;
    property Hints: TStrings read FHints write SetHints;
    property ParentCtl3D;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick: TNavClickEvent read FOnNavClick write FOnNavClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnResize;
  end;

  TNavButton = class(TSpeedButton)
  private
    FIndex      : TNavigateBtn;
    FNavStyle   : TNavButtonStyle;
    FRepeatTimer: TTimer;
    procedure TimerExpired(Sender: TObject);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    destructor Destroy; override;
    property NavStyle: TNavButtonStyle read FNavStyle write FNavStyle;
    property Index : TNavigateBtn read FIndex write FIndex;
  end;


implementation

uses Dialogs, TypInfo;

{$R QDB.RES} { contains the string resources for error messages etc. }

{ the codes for the messages in QDB.RES }
const
   SMissing         = 1;
   SCorrupt         = 2;
   SDoorOpen        = 3;
   SReadOnly        = 4;
   STooMany         = 5;
   SShareError      = 6;
   SDiskFull        = 7;
   SUnknownError    = 8;
   SIndexAdd        = 9;
   SCannotCopy      = 10;
   SDataAdd         = 11;
   SDuplicateKey    = 12;
   SSortedListError = 13;
   SOutOfBounds     = 14;
   STempFile        = 15;
   SNoFile          = 16;
   SFilter          = 17;

type
  TIndex = class    { each index object points to variable length data... }
    Pos: TFilePos;  { it's location in the data file                      }
    Len: TDataIndex;{ and it's length                                     }
    Res: longint;   { reserved for future use                             }
  end;

type
  TCache = class    { each cache object holds variable length data        }
    Buf: PChar;     { the data pointer                                    }
    Len: TDataIndex;{ it's length                                         }
    ID : Double;    { and when it was cached (millisecond resolution)     }
  end;

{$IFDEF WIN32}
{ TMyStringList -- TStringList with minor but essential changes        }
{ Copyright (c) 1995, 1996 Borland International. All Rights Reserved. }

procedure MyStringListError(SErr: integer);
begin
raise EQDBListError.CreateRes(SErr);
end;

destructor TMyStringList.Destroy;
begin
  FOnChange := nil;
  FOnChanging := nil;
  if FCount <> 0 then Finalize(FList^[0], FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TMyStringList.Add(const S: string): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(S, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: MyStringListError(SDuplicateKey);
      end;
  InsertItem(Result, S);
end;

procedure TMyStringList.Changed;
begin
if (FUpdateCount = 0) and Assigned(FOnChange) then FOnChange(Self);
end;

procedure TMyStringList.Changing;
begin
if (FUpdateCount = 0) and Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TMyStringList.Clear;
begin
  if FCount <> 0 then
    begin
    Changing;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
    Changed;
  end;
end;

procedure TMyStringList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then MyStringListError(SOutOfBounds);
  Changing;
  Finalize(FList^[Index]);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TStringItem));
  Changed;
end;

procedure TMyStringList.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) or
    (Index2 < 0) or (Index2 >= FCount) then MyStringListError(SOutOfBounds);
  Changing;
  ExchangeItems(Index1, Index2);
  Changed;
end;

procedure TMyStringList.ExchangeItems(Index1, Index2: Integer);
var
  Temp: Integer;
  Item1, Item2: PStringItem;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];
  Temp := Integer(Item1^.FString);
  Integer(Item1^.FString) := Integer(Item2^.FString);
  Integer(Item2^.FString) := Temp;
  Temp := Integer(Item1^.FObject);
  Integer(Item1^.FObject) := Integer(Item2^.FObject);
  Integer(Item2^.FObject) := Temp;
end;

function TMyStringList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := AnsiCompareText(FList^[I].FString, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

function TMyStringList.Get(Index: Integer): string;
begin
  if (Index < 0) or (Index >= FCount) then MyStringListError(SOutOfBounds);
  Result := FList^[Index].FString;
end;

function TMyStringList.GetCapacity: integer;
begin
result:= FCapacity;
end;

function TMyStringList.GetCount: Integer;
begin
  Result := FCount;
end;

function TMyStringList.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then MyStringListError(SOutOfBounds);
  Result := FList^[Index].FObject;
end;

procedure TMyStringList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 8 then Delta := 16 else
    if FCapacity > 4 then Delta := 8 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TMyStringList.IndexOf(const S: string): Integer;
begin
  if not Sorted then Result := inherited IndexOf(S) else
    if not Find(S, Result) then Result := -1;
end;

procedure TMyStringList.Insert(Index: Integer; const S: string);
begin
  if Sorted then MyStringListError(SSortedListError);
  if (Index < 0) or (Index > FCount) then MyStringListError(SOutOfBounds);
  InsertItem(Index, S);
end;

procedure TMyStringList.InsertItem(Index: Integer; const S: string);
begin
  Changing;
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TStringItem));
  with FList^[Index] do
  begin
    Pointer(FString) := nil;
    FObject := nil;
    FString := S;
  end;
  Inc(FCount);
  Changed;
end;

procedure TMyStringList.Put(Index: Integer; const S: string);
begin
  if Sorted then MyStringListError(SSortedListError);
  if (Index < 0) or (Index >= FCount) then MyStringListError(SOutOfBounds);
  Changing;
  FList^[Index].FString := S;
  Changed;
end;

procedure TMyStringList.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then MyStringListError(SOutOfBounds);
  Changing;
  FList^[Index].FObject := AObject;
  Changed;
end;

procedure TMyStringList.QuickSort(L, R: Integer);
var
  I, J: Integer;
  P: string;
begin
  repeat
    I := L;
    J := R;
    P := FList^[(L + R) shr 1].FString;
    repeat
      while AnsiCompareText(FList^[I].FString, P) < 0 do Inc(I);
      while AnsiCompareText(FList^[J].FString, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

procedure TMyStringList.SetCapacity(NewCapacity: integer);
begin
  ReallocMem(FList, NewCapacity * SizeOf(TStringItem));
  FCapacity := NewCapacity;
end;

procedure TMyStringList.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TMyStringList.SetUpdateState(Updating: Boolean);
begin
  if Updating then Changing else Changed;
end;

procedure TMyStringList.Sort;
begin
  if not Sorted and (FCount > 1) then
  begin
    Changing;
    QuickSort(0, FCount - 1);
    Changed;
  end;
end;

{ just reverses the order of the list }
procedure TMyStringList.Reverse;
var
  n,m: integer;
begin
if Sorted then exit;
n:=1;
m:=FCount;
while n < m do
  begin
  Exchange(n-1,m-1);
  inc(n);
  dec(m);
  end;
end;

{ this dangerous routine bypasses the actual sorting    }
{ method and just sets the Sorted property to true ...  }
{ if the list isn't actually sorted you are in a mess!  }
procedure TMyStringList.PretendSort;
begin
FSorted:=true;
end;
{$ELSE}
function TMyStringList.GetCapacity: integer;
begin
Result:=0;
end;

procedure TMyStringList.SetCapacity(NewCapacity: integer);
begin
end;
{$ENDIF}

{ derived lists }

destructor TIndexList.Destroy;
begin
EmptyAndClear;
inherited Destroy;
end;

procedure TIndexList.EmptyAndClear;
begin
while Count > 0 do
  begin
  TIndex(Objects[0]).Free; { free the TIndex objects     }
  Delete(0);               { free the TStringList object }
  end;
end;

destructor TCacheList.Destroy;
begin
EmptyAndClear;
inherited Destroy;
end;

procedure TCacheList.EmptyAndClear;
begin
while Count > 0 do
  begin
  FreeMem(TCache(Objects[0]).buf,TCache(Objects[0]).Len);
  TCache(Objects[0]).Free;         { free the TCache object      }
  Delete(0);                       { free the TStringList object }
  end;
end;

{ Allocates memory for a buffer -- first tries to get      }
{ TryThisBig bytes but if not keeps halving the size until }
{ a block can be allocated. The actual amount allocated    }
{ is returned as Result.                                   }
function GetBuffer(var buffer: pointer; TryThisBig: longint): longint;
var
  ok: boolean;
begin
if TryThisBig < 1024 then TryThisBig:=1024;
{$IFNDEF WIN32}
if TryThisBig > MaxInt then TryThisBig:=MaxInt;
{$ENDIF}
ok:=false;
while not ok do
  begin
  try
    GetMem(buffer,TryThisBig);
    ok:=true;
  except
    on EOutOfMemory do TryThisBig:=TryThisBig div 2;
  end;
  end;
Result:=TryThisBig;
end;


{ TBFStream methods }

constructor TBFStream.Create(const FileName: string; Mode: word; BufferSize: longint);
begin
inherited Create(FileName ,Mode);
if BufferSize = -1 then
  begin
  BufferSize:=inherited Seek(0,2);
  inherited Seek(0,0);
  end;
{need to make sure we have zero byte at the end of the buffer}
buflen:=GetBuffer(pointer(buffer),BufferSize)-1;
isdirty:=False;
ResetBuffer;
end;

destructor TBFStream.Destroy;
begin
ResetBuffer;
FreeMem(buffer,buflen+1);
inherited Destroy;
end;

procedure TBFStream.ResetBuffer;
begin
if isdirty then
  begin
  Write(Buffer^,bufferpos - buffer);
  isdirty:=false;
  end;
FillChar(buffer^,buflen+1,#0);
bufferpos:=buffer + buflen;
bytesread:=buflen;
end;

procedure TBFStream.AdjustBuffer;
begin
if isdirty then
  begin
  Write(Buffer^,bufferpos - buffer);
  isdirty:=false;
  end;
FillChar(buffer^,buflen+1,#0);
inherited Seek(bufferpos-buffer-bytesread,1);
end;

function TBFStream.Seek(offset: longint; origin: word): longint;
begin
ResetBuffer;
Result:=inherited Seek(offset,origin);
end;

function TBFStream.GetLongint(var L: longint): boolean;
begin
Result:=false;
if (bufferpos - buffer + SizeOf(L) > bytesread) then
  begin
  if (bytesread < buflen) then
    Exit
  else
    begin
    AdjustBuffer;
    bufferpos:=buffer;
    bytesread:=Read(buffer^,buflen);
    Result:=GetLongint(L);
    end;
  end
else
  begin
  Move(bufferpos^,L,SizeOf(L));
  Inc(bufferpos,SizeOf(L));
  Result:=true;
  end;
end;

procedure TBFStream.PutLongint(const L: longint);
begin
if (bufferpos - buffer + SizeOf(L) > buflen) then
  begin
  ResetBuffer;
  bufferpos:=buffer;
  end;
Move(L,bufferpos^,SizeOf(L));
Inc(bufferpos,SizeOf(L));
isdirty:=true;
end;

function TBFStream.GetKey(var K: TKey): boolean;
begin
Result:=false;
if (bufferpos - buffer + StrLen(bufferpos) + 1 > bytesread) then
  begin
  if (bytesread < buflen) then
    Exit
  else
    begin
    AdjustBuffer;
    bufferpos:=buffer;
    bytesread:=Read(buffer^,buflen);
    Result:=GetKey(K);
    end;
  end
else
  begin
  K:=StrPas(bufferpos);
  Inc(bufferpos,Length(K)+1);
  Result:=true;
  end;
end;

procedure TBFStream.PutKey(const K: TKey);
begin
if (bufferpos - buffer + Length(K) + 1 > buflen) then
  begin
  ResetBuffer;
  bufferpos:=buffer;
  end;
StrPCopy(bufferpos,K);
Inc(bufferpos,Length(K)+1);
isdirty:=true;
end;



{$IFNDEF WIN32}
const
  MAX_PATH = 255;
{$ENDIF}

{ returns a unique name for a temp file }
function TempFileName(prefix: string): string;
const
  tmpstrlen = MAX_PATH;
var
  tmpname    : PChar;
  tmppath    : PChar;
  {$IFNDEF WIN32}
  dummy      : PChar;
  {$ENDIF}
begin
Result:='';
GetMem(tmpname,tmpstrlen);
try
  FillChar(tmpname^,tmpstrlen,#0);
  {$IFDEF WIN32}
  GetMem(tmppath,tmpstrlen);
  try
    FillChar(tmppath^,tmpstrlen,#0);
    GetTempPath(tmpstrlen,tmppath);
    GetTempFileName(tmppath,PChar(prefix),0,tmpname);
    Result:=tmpname;
  finally
    FreeMem(tmppath,tmpstrlen);
  end;
  {$ELSE}
  GetMem(dummy,length(prefix)+1);
  try
    GetTempFileName(#0,StrPCopy(dummy,prefix),0,tmpname);
    Result:=StrPas(tmpname);
  finally
    FreeMem(dummy,length(prefix)+1);
  end;
  {$ENDIF}
finally
  FreeMem(tmpname,tmpstrlen);
end;
end;

{ TTempBFStream methods }

const
  TmpFilePrefix = 'QDT';

constructor TTempBFStream.Create(const OldFileName: string);
begin
TmpFileName:=TempFileName(TmpFilePrefix);
if TmpFileName = '' then raise EQDBIndexError.CreateResFmt(STempFile,[FOldFileName]);
inherited Create(TmpFileName,fmCreate or fmShareExclusive,32*1024);
FOldFileName:=OldFileName;
end;

destructor TTempBFStream.Destroy;
begin
inherited Destroy;
if TmpFileName <> '' then
  begin
  SysUtils.DeleteFile(FOldFileName);
  RenameFile(TmpFileName,FOldFileName);
  end;
TmpFileName:='';
FOldFileName:='';
end;


{ TQDB private methods }

procedure TQDB.FileError(ErrCode: Integer; SDefault: integer);
var
  SErr: integer;
begin
case ErrCode of
-1  : SErr:=SDefault;
-2  : SErr:=SMissing;
-3  : SErr:=SDoorOpen;
-4  : SErr:=STooMany;
-5  : SErr:=SShareError;
-101: SErr:=SDiskFull;
else
  raise EQDBFileError.CreateResFmt(SUnknownError,[IntToStr(ErrCode)]);
end;
raise EQDBFileError.CreateResFmt(SErr,[FFileName]);
end;

procedure TQDB.IndexError(SErr: integer);
begin
raise EQDBIndexError.CreateResFmt(SErr,[FFileName]);
end;

function TQDB.GetFileName: TQDBFileName;
begin
Result:= FFileName;
end;

{ setting the FileName property loads and unloads QDB files }
procedure TQDB.SetFileName(value: string);
begin
try
if value <> '' then value:=ExpandFilename(value);
if ExpandFileName(FFileName) <> ExpandFileName(value) then
  begin
  if not (csDesigning in ComponentState) and (FFileName <> '') then CloseQDB;
  if value = '' then
    FFileName:=value
  else
    begin
    FFileName:=ExpandFileName(value);
    if ExtractFileExt(FFilename) = '' then FFileName:=ChangeFileExt(FFileName,'.QDB');
    end;
  if not (csDesigning in ComponentState) and (FFileName <> '') then OpenQDB;
  FileAssigned; { trigger event }
  end;
except
  FFileName := ''; { if anything goes wrong ... }
  raise;
end;
end;

{ does all the work of moving about the index }
procedure TQDB.SetItemIndex(value: TItemIndex);
begin
if Index.Count = 0 then
  NoData                 { trigger event if empty }
else if (value < 0) or (value > Index.Count -1) then
  OutOfBounds            { trigger event if illegal move }
else
  begin
  FItemIndex:=value;                 { new index position }
  FKey:=GetKey(FItemIndex);          { updated properties }
  FItemSize:=GetItemSize(FItemIndex);{    "        "      }
  if (FUpdating = 0) then Navigate;  { trigger event when index pos changes }
  end
end;

{ defines the wild card chars for the pattern matching }
procedure TQDB.SetMatchChars(front: char; back: char);
begin
if front <> #0 then FFrontWild:=front;
if back <> #0 then FBackWild:=back;
end;

{ used by the filtering system  -- if you wanted a more }
{ sophisticated match you could override this function. }
function TQDB.Matches(key : TKey; pattern : TKey) : boolean;
var
  IsMatch    : boolean;
  LeftPos,
  RightPos,
  PatternPosF,
  PatternPosL,
  LenP,
  LastPos    : byte;
begin
LenP := Length(pattern);
LeftPos := Pos(FFrontWild, pattern);
if LeftPos = 1 then
   begin
   dec(LenP);
   pattern:=Copy(pattern,2,LenP);
   end;
RightPos:= Pos(FBackWild, pattern);
if RightPos = LenP then
   begin
   dec(LenP);
   pattern:=Copy(pattern,1,LenP);
   end;
if LenP = 0 then
   begin
   PatternPosF := 1;
   PatternPosL := 1;
   end
else
   begin
   PatternPosF := Pos(pattern,key);
   PatternPosL:=PatternPosF;
   if PatternPosL <> 0 then
     begin
     LastPos:=PatternPosL;
     while LastPos <> 0 do
       begin
       LastPos:=Pos(pattern, Copy(key,PatternPosL+1,Length(key)));
       PatternPosL:=PatternPosL+LastPos;
       end;
     end;
   end;
IsMatch := PatternPosF <> 0;
if IsMatch and (LeftPos = 0) and (PatternPosF <> 1) then
   begin
   IsMatch :=false;
   end;
if IsMatch and (RightPos = 0) and (PatternPosL <> (Length(key)-LenP+1)) then
   begin
   IsMatch := false;
   end;
Result := IsMatch;
end;

{ sets a filter to restrict movement }
procedure TQDB.SetFilter(value: TKey);
begin
Filtered:=(value <> ''); { we use this flag a lot elsewhere }
FFilter:=value;
end;

{ sets the upper limit on the cache }
procedure TQDB.SetCacheSize(value: longint);
begin
if value <> FCacheSize then
  begin
  FlushCache;
  FCacheSize:=value;
  end;
end;

{ gets the percentage of accesses that hit the cache }
function TQDB.GetCacheFrequency: integer;
begin
if FCacheTotal <> 0 then
  Result:=Round((100.0 * FCacheHits) / FCacheTotal)
else
  Result:=0;
end;

{ resets cache-hit counting -- notice that whatever the }
{ value the result is the same.                         }
procedure TQDB.SetCacheFrequency(value: integer);
begin
FCacheHits:=0;
FCacheTotal:=0;
end;

{ Usually the memory allocated for the index grows every few     }
{ new items or so, which can be time-consuming with all the      }
{ moving of memory blocks, etc. Instead PrepareToAdd allocates   }
{ the memory required in one chunk which is much quicker and     }
{ dramatically reduces memory fragmentation. Getting TStringList }
{ to expose its Set/GetCapacity methods has meant delving into   }
{ the VCL source (Copyright Borland Int'nl. All Rights Reserved.)}
procedure TQDB.PrepareToAdd(numberofitems: integer);
var
  CurrentCapacity: integer;
begin
CurrentCapacity:=Index.Count; { GetCapacity; }
Index.SetCapacity(CurrentCapacity + numberofitems);
end;

{ fetches the key for the given item }
function TQDB.GetKey(value: TItemIndex): TKey;
begin
if Index.Count < 1 then
  begin
  Result:='';
  NoData;
  end
else
  Result:=Index.Strings[value];
end;

{ fetches the size of the given item }
function TQDB.GetItemSize(value: TItemIndex): TDataIndex;
begin
if Index.Count < 1 then
  begin
  Result:=0;
  NoData;
  end
else
  Result:=TIndex(Index.Objects[value]).Len;
end;

{ Copies the in-memory list of keys to the Keys parameter. }
{ Items that are filtered out are not included.            }
procedure TQDB.AssignKeyList(Keys: TStrings);
var
  ThisItem      : TItemIndex;
  ThisKey       : TKey;
  WasSorted     : boolean;
begin
TStringList(Keys).Clear;
{ TStrings has no sort method but TStringList does }
if Keys is TStringList then
  begin
  WasSorted:=TStringList(Keys).Sorted;
  TStringList(Keys).Sorted:=false;
  end;
MonitorSetup(Index.Count, prKeyList);
for ThisItem:=1 to Index.Count do
  begin
  MonitorUpdate(ThisItem);
  if Filtered then
    begin
    ThisKey:=GetKey(ThisItem - 1);
    if Matches(ThisKey,FFilter) then
      TStringList(Keys).Add(ThisKey);
    end
  else
    TStringList(Keys).Add(GetKey(ThisItem - 1));
  end;
if Keys is TStringList then TStringList(Keys).Sorted:=WasSorted;
MonitorZero;
end;

{ next two methods are used to make read-only properties }
{ visible in Object Inspector                            }

procedure TQDB.SetDummyAuthor(value: string40);
begin
{ does nothing }
end;

procedure TQDB.SetDummyVersion(value: string05);
begin
{ does nothing }
end;

{ when a new QDB file is needed... }
procedure TQDB.CreateQDB;
const
  sig1 : array [0..3] of char = ('Q', 'D', 'B', #0);
  sig2 : array [0..3] of char = ('Q', 'I', 'X', #0);
  var
  f   : TFileHandle;
  sz1,
  sz2 : longint;
  zero: byte;
  bigzero: longint;
begin
zero := 0;  { to null-terminate the index block       }
bigzero:=0; { number of items in the index = 0        }
f:=FileCreate(FileName);
if f < 0 then FileError(f,0);
try   { protect file f }
  FileWrite(f,sig1,SizeOf(sig1));
  sz1:=0;
  FileWrite(f,sz1,SizeOf(sz1));
  FileWrite(f,sig2,SizeOf(sig2));
  sz2:=SizeOf(bigzero) + SizeOf(zero);
  FileWrite(f,sz2,SizeOf(sz2));
  FileWrite(f,bigzero,SizeOf(BigZero));
  FileWrite(f,zero,SizeOf(zero));
finally
  FileClose(f);
end;
end;

{ This seems like a good place to document the QDB file format ... }
{                                                                  }
{ 4 bytes -- 'Q','D','B',#0                                        }
{ 4 bytes -- longint(n1) = size of data block                      }
{ n1 bytes of data                                                 }
{ 4 bytes -- 'Q','I','X',#0                                        }
{ 4 bytes -- longint(n2) = size of index block                     }

{ Split takes a QDB file and, testing it for integrity,  }
{ splits it into two working files, one the item data,   }
{ the other the index data. These are the files that the }
{ QDB component uses internally. The original QDB file   }
{ doesn't get reconstituted until the Splice method.     }
procedure TQDB.Split;
var
  n          : integer;
  f,f1,f2    : TFileHandle;
  buffer     : pointer;
  sig        : array [0..3] of char;
  sz1,sz2    : longint;
  BytesRead  : TFilePos;
  buflen     : longint;
begin
{ we get the age of the file when it was opened }
FFileAge:=FileAge(FFileName);
f:=FileOpen(FFileName,fmOpenReadWrite);
if f < 0 then FileError(f,0);
try  { protect file f }
  FileRead(f,sig,SizeOf(sig));
  if StrPas(sig) <> 'QDB' then FileError(-1,SCorrupt);
  FileRead(f,sz1,SizeOf(sz1));
  buflen:=GetBuffer(buffer, sz1);
  try   { protect memory buffer }
    QDBFileName:=TempFileName('QDB');
    f1:=FileCreate(QDBFileName);
    if f1 < 0 then FileError(f1,0);
    try   { protect file f1 }
      for n:= 1 to (sz1 div buflen) do
        begin
        BytesRead:=FileRead(f,buffer^,buflen);
        FileWrite(f1,buffer^,BytesRead);
        end;
      BytesRead:=FileRead(f,buffer^,sz1 mod buflen);
      FileWrite(f1,buffer^,BytesRead);
    finally
      FileClose(f1);
    end;
  finally
    FreeMem(buffer,buflen);
  end;
  FileRead(f,sig,SizeOf(sig));
  if StrPas(sig) <> 'QIX' then FileError(-1,SCorrupt);
  FileRead(f,sz2,SizeOf(sz2));
  buflen:=GetBuffer(buffer, sz2);
  try   { protect memory buffer }
    { use same 'random' name as for QDBFileName }
    QIXFileName:=ExtractFilePath(QDBFileName) + 'QIX' + Copy(ExtractFileName(QDBFileName),4,MAX_PATH);
    f2:=FileCreate(QIXFileName);
    if f2 < 0 then FileError(f2,0);
    try   { protect file f2 }
      for n:= 1 to (sz2 div buflen) do
        begin
        BytesRead:=FileRead(f,buffer^,buflen);
        FileWrite(f2,buffer^,BytesRead);
        end;
      BytesRead:=FileRead(f,buffer^,sz2 mod buflen);
      FileWrite(f2,buffer^,BytesRead);
    finally
      FileClose(f2);
    end;
  finally
    FreeMem(buffer,buflen);
  end;
finally
  FileClose(f);
end;
end;

{ Splices the two working files back into a single QDB file,   }
{ takingcare not to overwrite the original if it has been used }
{ by another program.                                          }
procedure TQDB.Splice;
const
  sig1 : array [0..3] of char = ('Q', 'D', 'B', #0);
  sig2 : array [0..3] of char = ('Q', 'I', 'X', #0);
var
  n           : integer;
  f,f1,f2     : TFileHandle;
  buffer      : pointer;
  sz1,sz2     : longint;
  BytesRead   : TFilePos;
  buflen      : longint;
  CanOverWrite: boolean;
begin
if ReadOnly then Exit;
{ check if the file has been used since we opened it -- }
{ if FFileAge = 0 it means we are doing a SaveAs...     }
if (FFileAge <> 0) and (FileAge(FFileName) <> FFileAge) then
  begin
  { it has so we need to ask if we can save our stuff over it }
  CanOverWrite:=false;           { be conservative }
  AboutToOverWrite(CanOverWrite);
  if not CanOverWrite then
    repeat
    { since we can't overwrite the original we have   }
    { to look for a unique derivative of the filename }
    FFileName:=ExtractFilePath(FFileName) + '1.' + ExtractFileName(FFilename);
    until not FileExists(Filename);
    { then we can go ahead as normal }
  end;
f:=FileCreate(FFileName);
if f < 0 then FileError(f,0);
try   { protect file f }
  FileWrite(f,sig1,SizeOf(sig1));
  f1:=FileOpen(QDBFileName,fmOpenReadWrite or fmShareExclusive);
  if f1 < 0 then FileError(f1,0);
  try   { protect file f1 }
    sz1:=FileSeek(f1,0,2);
    FileWrite(f,sz1,SizeOf(sz1));
    FileSeek(f1,0,0);
    buflen:=GetBuffer(buffer,sz1);
    try   { protect memory buffer }
      for n:= 1 to (sz1 div buflen) do
        begin
        BytesRead:=FileRead(f1,buffer^,buflen);
        FileWrite(f,buffer^,BytesRead);
        end;
      BytesRead:=FileRead(f1,buffer^,sz1 mod buflen);
      FileWrite(f,buffer^,BytesRead);
    finally
      FreeMem(buffer, buflen);
    end;
  finally
    FileClose(f1);
  end;
  FileWrite(f,sig2,SizeOf(sig2));
  f2:=FileOpen(QIXFileName,fmOpenReadWrite or fmShareExclusive);
  if f2 < 0 then FileError(f2,0);
  try   { protect file f2 }
    sz2:=FileSeek(f2,0,2);
    FileWrite(f,sz2,SizeOf(sz2));
    FileSeek(f2,0,0);
    buflen:=GetBuffer(buffer,sz2);
    try  { protect memory buffer }
      for n:= 1 to (sz2 div buflen) do
        begin
        BytesRead:=FileRead(f2,buffer^,buflen);
        FileWrite(f,buffer^,BytesRead);
        end;
      BytesRead:=FileRead(f2,buffer^,sz2 mod buflen);
      FileWrite(f,buffer^,BytesRead);
    finally
      FreeMem(buffer, buflen);
    end;
  finally
    FileClose(f2);
  end;
finally
  FileClose(f);
  FFileAge:=FileAge(FFileName);
end;
end;

{ returns the name of an orphaned working file }
function TQDB.FileToRecover: string;
var
  tmppath       : PChar;
  placetolook   : string;
  F             : TSearchRec;
  OldQDBFileName: string;
  handle        : integer;
begin
Result:='';
GetMem(tmppath,MAX_PATH);
try
  FillChar(tmppath^,MAX_PATH,#0);
  {$IFDEF WIN32}
  GetTempPath(MAX_PATH,tmppath);
  placetolook:=tmppath;
  {$ELSE}
  GetTempFileName(#0,'RRM',0,tmppath);
  placetolook:=ExtractFilePath(StrPas(tmppath));
  {$ENDIF}
finally
  Freemem(tmppath,MAX_PATH);
end;
if FindFirst(placetolook + 'QDB*.tmp', 0, F) = 0 then
  begin
  OldQDBFileName:=placetolook + F.Name;
  SysUtils.FindClose(F);
  { check if in use at the moment or if genuine orphan ... }
  handle:=FileOpen(OldQDBFileName,fmShareExclusive);
  if handle < 0 then exit;
  FileClose(handle);
  Result:=OldQDBFileName;
  end;
end;

{ is there at least one ... }
function TQDB.OrphanToRecover: boolean;
begin
Result:=(FileToRecover <> '');
end;

{ Checks to see if there are temp working files still around }
{ that are not in use.                                       }
procedure TQDB.Recover(NewFileName: string);
var
  Remnant: string;
begin
if FFileName <> '' then exit; { only use if nothing open }
Remnant:=FileToRecover;
if Remnant = '' then exit; { and there is an orphan }
{ prepare to splice the working files together }
QDBFileName:=Remnant;
QIXFileName:=ExtractFilePath(Remnant) + 'QIX' + Copy(ExtractFileName(Remnant),4,MAX_PATH);
FFileAge:=0;
FFileName:=NewFileName;
Splice;
FFileName:='';
SysUtils.DeleteFile(QDBFileName);
SysUtils.DeleteFile(QIXFileName);
{ then open it up again }
Filename:=NewFileName;
end;

{ open a QDB file }
procedure TQDB.OpenQDB;
begin
try
  if not FileExists(FFileName) then CreateQDB;
  LoadIndex;
  FReady:=true;
except
  raise
end;
end;

{ close up a QDB file }
procedure TQDB.CloseQDB;
begin
SaveIndex;
FItemIndex:=0;
FKey:='';
FCount:=0;
FItemSize:=0;
FReadOnly:=false;
IsDirty:=false;
FReady:=false;
end;

{ Loading and saving the index are the most complex tasks in the unit }
{ Some words about the working file formats is in order:              }
{ The item file (QDBFile) consists of contiguous, variable-length,    }
{ blocks of data. Where one ends and the next begins is known only    }
{ to the index file (QIXFile). LoadIndex reads this data into         }
{ its Index list. The index file format is as follows:                }
{                                                                     }
{ 4 bytes -- longint(n) = number of items in the file                 }
{ n variable length blocks of the following structure                 }
{    null terminated string data = the key to an item                 }
{    4 bytes -- longint(n1) = the place in the item file              }
{    4 bytes -- longint(n2) = the size of the item                    }
{    4 bytes -- longint(n3) = reserved (should be zero)               }
{                                                                     }
procedure TQDB.LoadIndex;
var
  key      : TKey;
  rec      : TIndex;
  NumItems : longint;
  n        : longint;
begin
try  { except any error }
  Split;
  QDBFile:=FileOpen(QDBFileName,fmOpenReadWrite or fmShareExclusive);
  if QDBFile < 0 then FileError(QDBFile,0);
  QIXFile:=TBFStream.Create(QIXFileName,fmOpenRead or fmShareExclusive,-1);
  try  { protect file QIXFile }
    QIXFile.GetLongInt(NumItems);
    MonitorSetup(NumItems,prStart);
    Index.SetCapacity(NumItems);
    Index.Sorted:=false;  { quicker to sort later than add to a sorted list }
    FMaxItemSize:=0;
    for n:=1 to NumItems do
      begin
      QIXFile.GetKey(key);
      try
        rec:=TIndex.Create;
      except
        IndexError(SIndexAdd);
      end;
      QIXFile.GetLongint(rec.Pos);
      QIXFile.GetLongint(rec.Len);
      QIXFile.GetLongint(rec.Res);
      try
        FItemIndex:=Index.AddObject(key,rec);
      except
        rec.Free;
        Index.EmptyAndClear;
        Index.Sorted:=true;
        IndexError(SIndexAdd);
      end;
      if rec.Len > FMaxItemSize then FMaxItemSize:=rec.len;
      MonitorUpdate(n);
      end;
    {FReadOnly:=false;}
    FItemIndex:=0;
    FCount:=Index.Count;
    IsDirty:=false;
    {$IFDEF WIN32}
    { Note the fiddle coming next -- we don't sort the index }
    { because we know that the items were saved sorted but   }
    { in reverse order -- we just reverse the order again ...}
    Index.Reverse;
    { ... and pretend to sort, i.e., we just tell the index  }
    { it is sorted without actually sorting it               }
    Index.PretendSort;
    {$ELSE}
    Index.Sorted:=true;
    {$ENDIF}
    MonitorZero;
    if FCount > 0 then
      begin
      FKey:=GetKey(FItemIndex);
      FItemSize:=GetItemSize(FItemIndex);
      end
    else
      begin
      FKey:='';
      FItemSize:=0;
      end;
  finally
    QIXFile.Destroy;
  end;
except
  on EQDBIndexError do
    begin
    QDBFile:=0;
    {QIXFile:=0;}
    raise
    end;
  on EOutOfMemory do IndexError(SCorrupt);
  on ERangeError do IndexError(SCorrupt);
  on EQDBListError do IndexError(SCorrupt);
  else raise;
end;
end;

{ Saving the index is just as messy as loading it ... }
procedure TQDB.SaveIndex;
var
  t       : TIndex;
  n       : Integer;
  NumItems: longint;
  tmpfile : TTempBFStream;
begin
MonitorSetup(Index.Count,prFinish);
if ReadOnly then
  begin
  n:=0;
  while Index.Count > 0 do
    begin
    MonitorUpdate(n);
    Inc(n);
    TIndex(Index.Objects[Index.Count-1]).Free;
    Index.Delete(Index.Count-1);
    end;
  end
else
  begin
  tmpfile:=TTempBFStream.Create(QIXFileName);
  try  { protect file tmpfile }
    NumItems:=Index.Count;
    tmpfile.PutLongint(NumItems);
    n:=0;
    while Index.Count > 0 do
      begin
      MonitorUpdate(n);
      Inc(n);
      tmpfile.PutKey(Index.Strings[Index.Count-1]);
      t:=TIndex(Index.Objects[Index.Count-1]);
      tmpfile.PutLongint(t.Pos);
      tmpfile.PutLongint(t.Len);
      tmpfile.PutLongint(t.Res);
      t.Free;
      Index.Delete(Index.Count-1);
      end;
  finally
    tmpfile.Destroy;
  end;
  end;
FileClose(QDBFile);
IsDirty:=false;
Splice;
SysUtils.DeleteFile(QDBFileName);
SysUtils.DeleteFile(QIXFileName);
MonitorZero;
end;

{ Save commits the in-memory index to the working file }
procedure TQDB.Save;
var
  t       : TIndex;
  i,n     : Integer;
  NumItems: longint;
  tmpfile : TTempBFStream;
begin
if not FReady then IndexError(SNoFile);
if ReadOnly or not IsDirty then exit;
MonitorSetup(Index.Count,prSave);
tmpfile:=TTempBFStream.Create(QIXFileName);
try  { protect file tmpfile }
  NumItems:=Index.Count;
  tmpfile.PutLongint(NumItems);
  i:=Index.Count;
  n:=0;
  while i > 0 do
    begin
    MonitorUpdate(n);
    Inc(n);
    tmpfile.PutKey(Index.Strings[i-1]);
    t:=TIndex(Index.Objects[i-1]);
    tmpfile.PutLongint(t.Pos);
    tmpfile.PutLongint(t.Len);
    tmpfile.PutLongint(t.Res);
    dec(i);
    end;
finally
  tmpfile.Destroy;
end;
FileClose(QDBFile);
IsDirty:=false;
QDBFile:=FileOpen(QDBFileName,fmOpenReadWrite or fmShareExclusive);
if QDBFile < 0 then FileError(QDBFile,0);
MonitorZero;
end;

{ SaveAs first commits the in-memory index to the working file }
{ before writing a copy of index and items to a new QDB file   }
procedure TQDB.SaveAs(NewName: string);
begin
Save;
FileClose(QDBFile);
FFileName:=ExpandFileName(NewName);
if ExtractFileExt(FFilename) = '' then FFileName:=ChangeFileExt(FFileName,'.QDB');
FFileAge:=0;
Splice;
FFileAge:=FileAge(FFileName);
QDBFile:=FileOpen(QDBFileName,fmOpenReadWrite or fmShareExclusive);
if QDBFile < 0 then FileError(QDBFile,0);
IsDirty:=false;
end;

{ Re-organizes the working item file into index order }
{ eliminating and unreferenced items.                 }
procedure TQDB.Pack;
var
  tmpfile    : TFileHandle;
  DatBuf     : PChar;
  QLen       : TDataIndex;
  QPos       : TFilePos;
  TmpFileName: string;
  This, Init : TItemIndex;
begin
if not FReady then IndexError(SNoFile);
if ReadOnly then
  begin
  CannotChange;
  Exit;
  end;
Init:=FItemIndex;
TmpFileName:=TempFileName('QDB');
if TmpFileName = '' then IndexError(STempFile);
tmpfile:=FileCreate(TmpFileName);
if tmpfile < 0 then IndexError(STempFile);
try  { protect file tmpfile }
  MonitorSetup(Index.Count,prPack);
  for This:=1 to Index.Count do
    begin
    MonitorUpdate(This);
    QLen:=TIndex(Index.Objects[This-1]).Len;
    QPos:=TIndex(Index.Objects[This-1]).Pos;
    GetMem(DatBuf,QLen);
    try  { protect memory DatBuf }
      try  { catch file errors }
        TIndex(Index.Objects[This-1]).Pos:=FileSeek(tmpfile,0,2);
        FileSeek(QDBFile, QPos, 0);
        FileRead(QDBFile, DatBuf^, QLen);
      except
        IndexError(SCorrupt);
      end;
      FileWrite(tmpfile,DatBuf^,QLen);
    finally
      FreeMem(DatBuf,QLen);
    end;
    end;
finally
  FileClose(tmpfile);
end;
FileClose(QDBFile);
SysUtils.DeleteFile(QDBFileName);
RenameFile(TmpFileName,QDBFileName);
QDBFile:=FileOpen(QDBFileName,fmOpenReadWrite or fmShareExclusive);
if QDBFile < 0 then FileError(QDBFile,0);
ItemIndex:=Init;
MonitorZero;
end;

{ Erases a QDB file after asking confirmation }
procedure TQDB.Kill;
var
  OK        : boolean;
  FileToDel : string;
begin
if not FReady then IndexError(SNoFile);
OK:=false;
AboutToKill(OK);
if not OK then Exit;
if ReadOnly then
  begin
  CannotChange;
  Exit;
  end;
if FFileName <> '' then
  begin
  FileToDel:=FileName;
  FileName:='';
  SysUtils.DeleteFile(FileToDel);
  Killed;
  end;
end;

{ TQDB protected methods }

{ a pile of event handlers ... }

procedure TQDB.Navigate;
begin
if Assigned(FOnNavigate) then FOnNavigate(Self);
end;

procedure TQDB.Added;
begin
if Assigned(FOnAdded) then FOnAdded(Self);
end;

procedure TQDB.Deleted;
begin
if Assigned(FOnDeleted) then FOnDeleted(Self);
end;

procedure TQDB.Changed;
begin
if Assigned(FOnChanged) then FOnChanged(Self);
end;

procedure TQDB.Found;
begin
if Assigned(FOnFound) then FOnFound(Self);
end;

procedure TQDB.FileAssigned;
begin
if Assigned(FOnFileAssigned) then FOnFileAssigned(Self);
end;

procedure TQDB.OutOfBounds;
begin
if Assigned(FWarnOutOfBounds) then FWarnOutOfBounds(Self);
end;

procedure TQDB.NoData;
begin
if Assigned(FWarnNoData) then FWarnNoData(Self);
end;

procedure TQDB.CannotChange;
begin
if Assigned(FWarnReadOnly) then FWarnReadOnly(Self);
end;

procedure TQDB.CannotCache;
begin
if Assigned(FWarnCannotCache) then FWarnCannotCache(Self);
end;

procedure TQDB.SignalProgress(Percent: TPercentage; Kind: TProgressOrigin);
begin
if Assigned(FProgressUpdate) then FProgressUpdate(Self,Percent,Kind);
end;

procedure TQDB.AboutToKill(var OK: boolean);
begin
if Assigned(FBeforeKill) then FBeforeKill(Self,OK);
end;

procedure TQDB.AboutToOverWrite(var OK: boolean);
begin
if Assigned(FBeforeOverWrite) then FBeforeOverWrite(Self,OK);
end;

procedure TQDB.Killed;
begin
if Assigned(FOnKilled) then FOnKilled(Self);
end;


{ TQDB public methods }

{ at last the constructor ! }
constructor TQDB.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
try
  Index:=TIndexList.Create;
  Index.Sorted:=true;
  Index.Duplicates:=dupError;
  Cache:=TCacheList.Create;
  Cache.Sorted:=true;
  Cache.Duplicates:=dupError;
  CurrentCacheSize:=0;
  FCacheSize:=16*1024;  { default cache size of 16K }
  CacheFrequency:=0;
  FReady:=false;
  FItemIndex:=0;
  FKey:='';
  FCount:=Index.Count;
  FItemSize:=0;
  FMaxItemSize:=0;
  FReadOnly:=false;
  FFrontWild:='<';
  FBackWild:='>';
  FVersion:=FileVersion;
  FAuthor:=AuthorInfo;
  FUpdating:=0;
except
  Index.Free;
  Cache.Free;
  raise;
end;
end;

destructor TQDB.Destroy;
begin
if not (csDesigning in ComponentState) and (FFileName <> '') then CloseQDB;
Index.Free;
Cache.Free;
FItemIndex:=0;
inherited Destroy;
end;

{ four methods that navigate the index, setting the ItemIndex            }
{ property. Note that setting this property triggers further processing. }

procedure TQDB.FirstItem;
var
  This: TItemIndex;
begin
FBoF:=true;
FEoF:=false;
if Filtered then
  begin
  { not quick ... }
  This:=0;
  while (This < Index.Count) and not Matches(GetKey(This),FFilter) do
  Inc(This);
  ItemIndex:=This;
  end
else
  ItemIndex:=0;
end;

procedure TQDB.LastItem;
var
  This: TItemIndex;
begin
FEoF:=true;
FBoF:=false;
if Filtered then
  begin
  This:=Index.Count-1;
  while (This >= 0) and not Matches(GetKey(This),FFilter) do Dec(This);
  ItemIndex:=This;
  end
else
  ItemIndex:=Index.Count - 1;
end;

procedure TQDB.NextItem;
var
  This: TItemIndex;
begin
FEoF:=false;
FBoF:=false;
if Filtered then
  begin
  This:=FItemIndex;
  while (This+1 < Index.Count) and not Matches(GetKey(This+1),FFilter) do
    begin
    Inc(This);
    end;
  Inc(This);
  if This >= Index.Count then
    begin
    FEoF:=true;
    FBof:=false;
    end;
  ItemIndex:=This;
  end
else
  begin
  This:=FItemIndex;
  Inc(This);
  if This >= Index.Count then
    begin
    FEoF:=true;
    FBof:=false;
    end;
  ItemIndex:=This;
  end;
end;

procedure TQDB.PrevItem;
var
  This: TItemIndex;
begin
FBoF:=false;
FEoF:=false;
if Filtered then
  begin
  This:=FItemIndex;
  while (This-1 >= 0) and not Matches(GetKey(This-1),FFilter) do
    begin
    Dec(This);
    end;
  Dec(This);
  if This < 0 then
    begin
    FBoF:=true;
    FEoF:=false;
    end;
  ItemIndex:=This;
  end
else
  begin
  This:=FItemIndex;
  Dec(This);
  if This < 0 then
    begin
    FBoF:=true;
    FEoF:=false;
    end;
  ItemIndex:=This;
  end;
end;


{ methods to implement the item cache }

{ add an item to the cache }
procedure TQDB.EnCache(ItemPtr: pointer; ItemLen: TDataIndex);
var
  CacheRec: TCache;
begin
{ we don't try to cache items bigger than the limit }
if ItemLen > FCacheSize then exit;
{ make room for the new entry by removing as many old ones as needed }
while CurrentCacheSize + ItemLen > FCacheSize do UnCache;
try
  CacheRec:=TCache.Create;
except
  CannotCache;
  Exit;
end;
CacheRec.Len:=ItemLen;
try
  GetMem(CacheRec.Buf,CacheRec.Len);
except
  CacheRec.Free;
  CannotCache;
  Exit;
end;
CacheRec.ID:=GetTickCount;
Move(ItemPtr^,CacheRec.Buf^,CacheRec.Len);
try
  Cache.AddObject(Index.Strings[FItemIndex],CacheRec);
except
  FreeMem(CacheRec.Buf,CacheRec.Len);
  CacheRec.Free;
  CannotCache;
  Exit;
end;
Inc(CurrentCacheSize,CacheRec.Len);
end;

{ remove the current item from the cache }
procedure TQDB.DeCache;
var
  CacheN: integer;
begin
if Cache.Find(Index.Strings[FItemIndex],CacheN) then
  begin
  Dec(CurrentCacheSize,TCache(Cache.Objects[CacheN]).Len);
  FreeMem(TCache(Cache.Objects[CacheN]).Buf,TCache(Cache.Objects[CacheN]).Len);
  TCache(Cache.Objects[CacheN]).Free;
  Cache.Delete(CacheN);
  end;
end;

{ remove the oldest (least recently used) item from the cache }
procedure TQDB.UnCache;
var
  n,
  oldest: integer;
  this,
  age   : Double;
begin
if Cache.Count < 1 then exit;
{ if there is a cache we scan through for the oldest item }
oldest:=0;
age:=GetTickCount;
for n:= 0 to Cache.Count-1 do
  begin
  this:=TCache(Cache.Objects[n]).ID;
  if this < age then
    begin
    age:=this;
    oldest:=n;
    end;
  end;
{ then we remove that cache entry }
with TCache(Cache.Objects[oldest]) do
  begin
  Dec(CurrentCacheSize,Len);
  FreeMem(Buf,Len);
  Free;
  end;
Cache.Delete(oldest);
end;

{ clear the cache }
procedure TQDB.FlushCache;
begin
while Cache.Count > 0 do
  begin
  FreeMem(TCache(Cache.Objects[0]).buf,TCache(Cache.Objects[0]).Len);
  TCache(Cache.Objects[0]).Free;     { free the TCache object }
  Cache.Delete(0);                   { free the string object }
  end;
CurrentCacheSize:=0;
CacheFrequency:=0;
end;

{ accesses the current item }
procedure TQDB.GetItem(ItemPtr: pointer);
var
  ItemLen: TDataIndex;
  CacheN: integer;
begin
if not FReady then IndexError(SNoFile);
if Index.Count > 0 then
  begin
  try
    { first check if in cache }
    if Cache.Find(Index.Strings[FItemIndex],CacheN) then
      begin
      Move(TCache(Cache.Objects[CacheN]).Buf^,ItemPtr^,TCache(Cache.Objects[CacheN]).Len);
      TCache(Cache.Objects[CacheN]).ID:=GetTickCount; { update the time }
      Inc(FCacheHits);
      end
    else
      begin
      FileSeek(QDBFile, TIndex(Index.Objects[FItemIndex]).Pos,0);
      ItemLen:=FileRead(QDBFile, ItemPtr^, TIndex(Index.Objects[FItemIndex]).Len);
      EnCache(ItemPtr,ItemLen);
      end;
    Inc(FCacheTotal);
  except
    IndexError(SCorrupt);
  end;
  end
else
  NoData;
end;


{ three search methods ...  }

{ simple stringlist find }
function TQDB.ExactMatch(key: TKey): boolean;
var
  n: TItemIndex;
begin
Result:=Index.Find(key,n);
if Result then
  begin
  ItemIndex:=n;
  Found;
  end;
end;

{ looking for a near match }
function TQDB.CloseMatch(partialkey: TKey): boolean;
var
  n: TItemIndex;
begin
Result:=Index.Find(partialkey,n);
if (FItemIndex <> n) and (n < Index.Count) then
  begin
  ItemIndex:=n;
  Found;
  end;
end;

{ brute force search for a pattern -- not quick! }
function TQDB.PatternMatch(pattern: TKey): boolean;
var
  n: TItemIndex;
  k: TKey;
  t: TItemIndex;
begin
t:=FItemIndex;
Result:=false;
for n:=t+1 to Count do
  begin
  k:=GetKey(n-1);
  if Matches(k,pattern) then
    begin
    Result:=true;
    Break;
    end;
  end;
if Result and (t<> n-1) then
  begin
  ItemIndex:=n-1;
  Found;
  end;
end;

{ three methods to manipulate items }

{ add an item to the file }
procedure TQDB.AddItem(ItemPtr: pointer; ItemLen: TDataIndex; key: TKey);
var
  IndRec: TIndex;
begin
if not FReady then IndexError(SNoFile);
if ReadOnly then
  CannotChange
else
  begin
  try
    IndRec:=TIndex.Create;
  except
    IndexError(SIndexAdd);
  end;
  IndRec.Pos:=FileSeek(QDBFile,0,2);
  if IndRec.Pos < 0 then FileError(-1,SCorrupt);
  IndRec.Len:=ItemLen;
  try
    FItemIndex:=Index.AddObject(key,IndRec);
    FKey:=GetKey(FItemIndex);   { updated properties }
    FItemSize:=GetItemSize(FItemIndex);
  except
    IndRec.Free;
    IndexError(SDuplicateKey);
  end;
  if FileWrite(QDBFile,ItemPtr^,ItemLen) <> ItemLen then IndexError(SDataAdd);
  EnCache(ItemPtr,ItemLen);
  if ItemLen > FMaxItemSize then FMaxItemSize:=ItemLen;
  FCount:=Index.Count;
  IsDirty:=true;
  Added;
  if (FUpdating = 0) then Navigate;
  end;
end;

{ delete and item from the index -- need to Pack to get it }
{ out of the item file                                     }
procedure TQDB.DeleteItem;
begin
if not FReady then IndexError(SNoFile);
if ReadOnly then
  CannotChange
else
  if Index.Count < 1 then
    NoData
else
  begin
  DeCache;
  TIndex(Index.Objects[FItemIndex]).Free;
  Index.Delete(FItemIndex);
  if FItemIndex > 0 then FItemIndex:=FItemIndex - 1;
  ItemIndex:=FItemIndex;
  FCount:=Index.Count;
  IsDirty:=true;
  Deleted;
  end;
end;

{ Change the contents of the current item }
procedure TQDB.ChangeItem(ItemPtr: pointer; ItemLen: TDataIndex);
var
  thiskey : TKey;
  IndRec  : TIndex;
begin
if not FReady then IndexError(SNoFile);
if Index.Count < 1 then
  NoData
else if ReadOnly then
  CannotChange
else
  begin
  DeCache;
  thiskey:=GetKey(FItemIndex);
  TIndex(Index.Objects[FItemIndex]).Free;
  Index.Delete(FItemIndex);
  try
    IndRec:=TIndex.Create;
  except
    IndexError(SIndexAdd);
  end;
  IndRec.Pos:=FileSeek(QDBFile,0,2);
  if IndRec.Pos < 0 then FileError(-1,SDataAdd);
  IndRec.Len:=ItemLen;
  try
    FItemIndex:=Index.AddObject(key,IndRec);
  except
    IndRec.Free;
    raise
  end;
  if FileWrite(QDBFile,ItemPtr^,ItemLen) <> ItemLen then FileError(-1,SDataAdd);
  EnCache(ItemPtr,ItemLen);
  if ItemLen > FMaxItemSize then FMaxItemSize:=ItemLen;
  FItemSize:=GetItemSize(FItemIndex);
  IsDirty:=true;
  Changed;
  end;
end;


{ two methods for handling items as streams }

{ add a stream to the file }
procedure TQDB.AddStreamItem(stream: TStream; key: TKey);
var
  ItemPtr: pointer;
begin
if not FReady then IndexError(SNoFile);
stream.Seek(0,0);
GetMem(ItemPtr,stream.Size);
try
  stream.Read(ItemPtr^,stream.Size);
  AddItem(ItemPtr,stream.Size,key);
  stream.Seek(0,0);
finally
  FreeMem(ItemPtr,stream.Size);
end;
end;

{ retrieve an item from the file as a stream }
procedure TQDB.GetStreamItem(stream: TStream);
var
  ItemPtr: pointer;
begin
if not FReady then IndexError(SNoFile);
stream.Seek(0,0);
GetMem(ItemPtr,ItemSize);
try
  GetItem(ItemPtr);
  stream.Write(ItemPtr^,ItemSize);
  stream.Seek(0,0);
finally
  FreeMem(ItemPtr,ItemSize);
end;
end;

{ Two methods to control batch actions -- if FUpdating is     }
{ not zero the OnNavigate event is not triggered in the usual }
{ places. Since OnNavigate handlers usually update the screen }
{ this cuts out flicker and speed things up. In addition they }
{ also call the index's methods of the same name ...          }
procedure TQDB.BeginUpdate;
begin
inc(FUpdating);
Index.BeginUpdate;
end;

procedure TQDB.EndUpdate;
begin
Index.EndUpdate;
dec(FUpdating);
if (FUpdating = 0) then navigate;
end;

(* Progress monitoring stuff *)

procedure TQDB.MonitorSetup(const max: longint; const kind: TProgressOrigin);
begin
MonitorInterval:=(max div 10) + 1;
MonitorKind:=kind;
end;

procedure TQDB.MonitorZero;
begin
SignalProgress(0,MonitorKind);
end;

procedure TQDB.MonitorMax;
begin
SignalProgress(100,MonitorKind);
end;

procedure TQDB.MonitorUpdate(const n: longint);
begin
if n mod MonitorInterval = 0 then SignalProgress((10*n) div MonitorInterval,MonitorKind);
end;


{ TQDBFilenameProperty Editor }

Procedure TQDBFilenameProperty.Edit;
var
  QDBFileOpen: TOpenDialog;
begin
  QDBFileOpen:=TOpenDialog.Create(Application);
  QDBFileOpen.Filename:=GetValue;
  QDBFileOpen.Filter:=LoadStr(SFilter);
  QDBFileOpen.HelpContext:=0;
  QDBFileOpen.Options:=[ofPathMustExist];
  try
    if QDBFileOpen.Execute then SetValue(QDBFileOpen.Filename);
  finally
    QDBFileOpen.Free;
  end;
end;

function TQDBFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog{, paRevertable}];
end;


{ TDBNavigator }

{ Basically -- the TQDBNavigator component from DBCtrls but with      }
{ all the BDE stuff torn out and replaced with QDB stuff instead.     }
{ Portions of this code are Copyright Borland.                        }
{ Copyright (c) 1995, 1996 Borland International. All Rights Reserved.}

{$R QDBNav.res}

const
  InitRepeatPause = 400;  { pause before repeat timer (ms)        }
  RepeatPause     = 100;  { pause before hint window displays (ms)}
  SpaceSize       =  5;   { size of space between special buttons }

const
  SFirstRecord = 62019;
  SPrevRecord  = 62020;
  SNextRecord  = 62021;
  SLastRecord  = 62022;
  BtnStateName: array[TNavGlyph] of PChar = ('EN', 'DI');
  BtnTypeName : array[TNavigateBtn] of PChar = ('FIRST', 'PREV', 'NEXT', 'LAST');
  BtnHintId: array[TNavigateBtn] of Word = (SFirstRecord, SPrevRecord, SNextRecord, SLastRecord);

constructor TQDBNavigator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csFramed, csOpaque];
  FVisibleButtons := [nbFirst, nbPrev, nbNext, nbLast];
  FHints := TStringList.Create;
  InitButtons;
  BevelOuter := bvNone;
  BevelInner := bvNone;
  Width := 121;
  Height := 25;
  ButtonWidth := 0;
  FocusedButton := nbFirst;
end;

destructor TQDBNavigator.Destroy;
begin
FHints.Free;
inherited Destroy;
end;

procedure TQDBNavigator.InitButtons;
var
  I: TNavigateBtn;
  Btn: TNavButton;
  X: Integer;
  ResName: array [0..40] of char;
begin
  MinBtnSize := Point(20, 18);
  X := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    Btn := TNavButton.Create (Self);
    Btn.Index := I;
    Btn.Visible := I in FVisibleButtons;
    Btn.Enabled := True;
    Btn.SetBounds (X, 0, MinBtnSize.X, MinBtnSize.Y);
    StrFmt(ResName, 'qdb_%s', [BtnTypeName[I]]);
    Btn.Glyph.Handle := LoadBitmap(HInstance, ResName);
    Btn.NumGlyphs := 2;
    Btn.OnClick := _Click;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    Buttons[I] := Btn;
    X := X + MinBtnSize.X;
  end;
  InitHints;
  Buttons[nbPrev].NavStyle := Buttons[nbPrev].NavStyle + [nsAllowTimer];
  Buttons[nbNext].NavStyle  := Buttons[nbNext].NavStyle + [nsAllowTimer];
end;

procedure TQDBNavigator.InitHints;
var
  I: Integer;
  J: TNavigateBtn;
begin
  for J := Low(Buttons) to High(Buttons) do
    Buttons[J].Hint := LoadStr (BtnHintId[J]);
  J := Low(Buttons);
  for I := 0 to (FHints.Count - 1) do
  begin
    if FHints.Strings[I] <> '' then Buttons[J].Hint := FHints.Strings[I];
    if J = High(Buttons) then Exit;
    Inc(J);
  end;
end;

procedure TQDBNavigator.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
  InitHints;
end;

procedure TQDBNavigator.SetVisible(Value: TButtonSet);
var
  I: TNavigateBtn;
  W, H: Integer;
begin
  W := Width;
  H := Height;
  FVisibleButtons := Value;
  for I := Low(Buttons) to High(Buttons) do
    Buttons[I].Visible := I in FVisibleButtons;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

procedure TQDBNavigator.AdjustSize (var W: Integer; var H: Integer);
var
  Count: Integer;
  MinW: Integer;
  I: TNavigateBtn;
  Space, Temp, Remain: Integer;
  X: Integer;
begin
  if (csLoading in ComponentState) then Exit;
  if Buttons[nbFirst] = nil then Exit;

  Count := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Inc(Count);
    end;
  end;
  if Count = 0 then Inc(Count);

  MinW := Count * (MinBtnSize.X - 1) + 1;
  if W < MinW then
    W := MinW;
  if H < MinBtnSize.Y then
    H := MinBtnSize.Y;

  ButtonWidth := ((W - 1) div Count) + 1;
  Temp := Count * (ButtonWidth - 1) + 1;
  if Align = alNone then
    W := Temp;

  X := 0;
  Remain := W - Temp;
  Temp := Count div 2;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Space := 0;
      if Remain <> 0 then
      begin
        Dec (Temp, Remain);
        if Temp < 0 then
        begin
          Inc (Temp, Count);
          Space := 1;
        end;
      end;
      Buttons[I].SetBounds (X, 0, ButtonWidth + Space, Height);
      Inc (X, ButtonWidth - 1 + Space);
    end
    else
      Buttons[I].SetBounds (Width + 1, 0, ButtonWidth, Height);
  end;
end;

procedure TQDBNavigator.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustSize (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

procedure TQDBNavigator.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;

  { check for minimum size }
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;

procedure TQDBNavigator._Click(Sender: TObject);
begin
  BtnClick (TNavButton (Sender).Index);
end;

procedure TQDBNavigator.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  OldFocus: TNavigateBtn;
begin
  OldFocus := FocusedButton;
  FocusedButton := TNavButton (Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then
      Exit;
  end
  else if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedButton) then
  begin
    Buttons[OldFocus].Invalidate;
    Buttons[FocusedButton].Invalidate;
  end;
end;

procedure TQDBNavigator.BtnClick(Index: TNavigateBtn);
begin
  if (FQDB <> nil) then
  begin
      case Index of
        nbPrev: FQDB.PrevItem;
        nbNext: FQDB.NextItem;
        nbFirst: FQDB.FirstItem;
        nbLast: FQDB.LastItem;
      end;
      with FQDB do
        begin
        Buttons[nbFirst].Enabled:= not BoF;
        Buttons[nbPrev].Enabled:= not BoF;
        Buttons[nbNext].Enabled:=not EoF;
        Buttons[nbLast].Enabled:=not EoF;
        end;
  end;
  if not (csDesigning in ComponentState) and Assigned(FOnNavClick) then
    FOnNavClick(Self, Index);
end;

procedure TQDBNavigator.WMSetFocus(var Message: TWMSetFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TQDBNavigator.WMKillFocus(var Message: TWMKillFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TQDBNavigator.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewFocus: TNavigateBtn;
  OldFocus: TNavigateBtn;
begin
  OldFocus := FocusedButton;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus < High(Buttons) then
            NewFocus := Succ(NewFocus);
        until (NewFocus = High(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus > Low(Buttons) then
            NewFocus := Pred(NewFocus);
        until (NewFocus = Low(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_SPACE:
      begin
        if Buttons[FocusedButton].Enabled then
          Buttons[FocusedButton].Click;
      end;
  end;
end;

procedure TQDBNavigator.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TQDBNavigator.Loaded;
var
  W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  InitHints;
end;

{ TNavButton }

destructor TNavButton.Destroy;
begin
  if FRepeatTimer <> nil then
    FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TNavButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if nsAllowTimer in FNavStyle then
  begin
    if FRepeatTimer = nil then
      FRepeatTimer := TTimer.Create(Self);

    FRepeatTimer.OnTimer := TimerExpired;
    FRepeatTimer.Interval := InitRepeatPause;
    FRepeatTimer.Enabled  := True;
  end;
end;

procedure TNavButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
                                  X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  if FRepeatTimer <> nil then
    FRepeatTimer.Enabled  := False;
end;

procedure TNavButton.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  if (FState = bsDown) and MouseCapture then
  begin
    try
      Click;
    except
      FRepeatTimer.Enabled := False;
      raise;
    end;
  end;
end;

procedure TNavButton.Paint;
var
  R: TRect;
begin
  inherited Paint;
  if (GetFocus = Parent.Handle) and
     (FIndex = TQDBNavigator (Parent).FocusedButton) then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then
      OffsetRect(R, 1, 1);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;

end.
