Unit RegFiles;

interface

uses Windows,SysUtils,Classes;

type
   TRootKey = (hkClassesRoot,hkCurrentUser,hkLocalMachine,
               hkUsers,hkCurrentConfig,hkDynData);

   ERegFileError = class(Exception);

   TRegFile = class(TComponent)
    private
      fRootKey:TRootKey;
      fKey:String;
      fOpenKey:HKEY;
      fLastError:DWORD;
      function MergeKey(AKey:String):String;
      procedure RegCheck(Rslt:DWORD);
      function RegDataSize(AValue:String; var RegType:DWORD):integer;
      function RegGetData(AValue:String; Buf:PByte; BufLen:DWORD):DWORD;
      procedure RegPutData(AValue:String;Buf:PByte; BufLen:DWORD; RegType:DWORD);
    protected
      function OpenKey(TheKey:String; MustExist:boolean):boolean; virtual;
      procedure CloseKey; virtual;
      function GetValue(Index:String):Variant;
      procedure SetValue(Index:String; TheValue:Variant);
      function GetValueDefault(Index:String; TheDefault:Variant):Variant;
    public
      constructor Create(AOwner:TComponent); override;
      function KeyExists(AKey:String):boolean;
      function ValueExists(AKey,AValue:String):boolean;
      procedure GetSubKeys(AKey:String; AList:TStrings);
      function HasSubKey(AKey:String):boolean;
      procedure GetValueNames(ASubKey:String; AList:TStrings);
      procedure DeleteKey(AKey:String);
      procedure DeleteValue(AKey:String; AValue:String);
      function RenameValue(AKey,OldName,NewName:String):boolean;
      function RenameKey(OldName,NewName:String):boolean;
      function BinaryDataSize(Index:String):integer;
      procedure GetBinaryData(Index:String; var Buff; BuffMax:integer);
      procedure PutBinaryData(Index:String; var Buff; BuffMax:integer);
      property LastError:DWORD read fLastError;
      property Value[Index:String]:Variant read GetValue write SetValue; default;
      property ValueDefault[Index:String; TheDefault:Variant]:Variant read GetValueDefault;
    published
      property RootKey:TRootKey read fRootKey write fRootKey default hkCurrentUser;
      property Key:String read fKey write fKey;
      end;

implementation

function ParseKey(var TheKey:String):String;
var
  i : integer;

  function NotASlash:boolean;
  begin
     if TheKey[i] = '\' then
        begin
        Result := false;
        if i<> 1 then
           if TheKey[i-1] = '\' then
              begin
              Result := true;
              Dec(i);
              end;
        end
     else
        Result := true;
  end;

begin
  i := Length(TheKey);
  While (i > 0) and (NotASlash) do Dec(i);
  Result := Copy(TheKey,1,i-1);
  if i <> 0 then
     begin
     System.Delete(TheKey,1,i);
     if Result[1] <> '\' then Result := '\'+Result;
     end;
end;

function Reg2Variant(RegType:DWORD; Buf:PByte; BufLen:DWORD):Variant;
begin
   Result := UnAssigned;
   case RegType of
     REG_SZ: Result := StrPas(Pchar(Buf));
     REG_DWORD:  Result := PInteger(Buf)^;
     REG_BINARY:
       if BufLen = SizeOf(TVarData) then
          Move(Buf^,TVarData(Result),Sizeof(TVarData));
   end;
end;

function Variant2Reg(V:Variant; var RType:DWORD; var TheBuff:PByte; var BufLen:DWORD):boolean;
begin
   Result := true;
   case VarType(V) of
     varString:
       begin
       BufLen := Length(V)+1;
       GetMem(TheBuff,BufLen);
       StrPCopy(PChar(TheBuff),V);
       RType := REG_SZ;
       end;
     varSmallInt,varInteger,varBoolean:
       begin
       BufLen := 4;
       GetMem(TheBuff,BufLen);
       PInteger(TheBuff)^ := V;
       RTYpe := REG_DWORD;
       end;
      varSingle,varDouble,varCurrency,varDate:
       begin
       BufLen := Sizeof(TVarData);
       GetMem(TheBuff,BufLen);
       Move(TVarData(V),TheBuff^,BufLen);
       RTYpe := REG_BINARY;
       end;
   else
     Result := false;
   end;
end;

constructor TRegFile.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  fRootKey := hkCurrentUser;
end;

procedure TRegFile.RegCheck(Rslt:DWORD);
begin
  fLastError := Rslt;
  if Rslt <> ERROR_SUCCESS then
     raise ERegFileError.Create(Format('Reg Error: %8.8X',[Rslt]));
end;

function TRegFile.OpenKey(TheKey:String; MustExist:boolean):boolean;
var
  lpdwDisposition : DWORD;
const
   KeyID : array[TRootKey] of HKEY =
     (HKEY_CLASSES_ROOT,HKEY_CURRENT_USER,HKEY_LOCAL_MACHINE,
      HKEY_USERS,HKEY_CURRENT_CONFIG,HKEY_DYN_DATA);
begin
  Result := true;
  if MustExist or (TheKey='') or (fRootKey=hkDynData) then
     begin
     Result := RegOpenKeyEx(KeyID[fRootKey],PChar(TheKey),0,
                            KEY_ALL_ACCESS,fOpenkey) = ERROR_SUCCESS;
     if not MustExist then
        if not Result then
           raise ERegFileError.Create('Can''t open a Key');
     end
  else
    { RegCheck(RegCreateKey(KeyID[fRootKey],PChar(TheKey),fOpenKey)); }
     RegCheck(RegCreateKeyEx(
      KeyID[fRootKey],    // handle of an open key
      PChar(TheKey),      // address of subkey name
      0,                  // reserved
      PChar(''),          // address of class string
      0,                  // special options flag
      KEY_ALL_ACCESS,     // desired security access
      nil,                // address of key security structure
      fOpenKey,           // address of buffer for opened handle
      @lpdwDisposition     // address of disposition value buffer
      ));
end;

procedure TRegFile.CloseKey;
begin
  if fOpenKey <> 0 then
     RegCheck(RegCloseKey(fOpenKey));
  fOpenKey := 0;
end;

function TRegFile.ValueExists(AKey,AValue:String):boolean;
var
  RegType:DWORD;
begin
  Result := OpenKey(MergeKey(AKey),true);
  if Result then
     begin
     Result := RegDataSize(AValue,RegType)>=0;
     CloseKey;
     end;
end;

function TRegFile.RegDataSize(AValue:String; var RegType:DWORD):integer;
var
  QRslt : DWORD;
  BufLen: DWORD;
begin
  { Get the buffer size }
  if AValue = '' then
     QRslt := RegQueryValueEx(fOpenKey,nil,nil,@RegType,nil,@BufLen)
  else
     QRslt := RegQueryValueEx(fOpenKey,Pchar(AValue),nil,@RegType,nil,@BufLen);
  if QRslt <> ERROR_SUCCESS then
     Result := -1
  else
     Result := BufLen;
end;

function TRegFile.RegGetData(AValue:String; Buf:PByte; BufLen:DWORD):DWORD;
var
  RegType: DWORD;
begin
   if AValue = '' then
      RegCheck(RegQueryValueEx(fOpenKey,nil,nil,@RegType,Buf,@BufLen))
   else
      RegCheck(RegQueryValueEx(fOpenKey,Pchar(AValue),nil,@RegType,Buf,@BufLen));
   Result := RegType;
end;

procedure TRegFile.RegPutData(AValue:String;Buf:PByte; BufLen:DWORD; RegType:DWORD);
begin
 if AValue = '' then
    RegCheck(RegSetValueEx(fOpenKey,nil,0,RegType,Buf,BufLen))
 else
    RegCheck(RegSetValueEx(fOpenKey,Pchar(AValue),0,RegType,Buf,BufLen));
end;

function TRegFile.GetValue(Index:String):Variant;
var
  RegType,BufLen:DWORD;
  Buf:PByte;
begin
   Result := NULL;
   OpenKey(fKey+ParseKey(Index),false);
   try
     { Get the buffer size }
     BufLen := RegDataSize(Index,RegType);
     if BufLen <= 0 then exit;
     GetMem(Buf,BufLen);
     try
      { Now get the data }
      RegType := RegGetData(Index,Buf,BufLen);
      Result := Reg2Variant(RegType,Buf,BufLen);
     finally
       FreeMem(Buf,BufLen);
       end;
   finally
     CloseKey;
     end;
end;

procedure TRegFile.SetValue(Index:String; TheValue:Variant);
var
  RegType,BufLen:DWORD;
  Buf:PByte;
begin
   OpenKey(fKey+ParseKey(Index),false);
   try
     if not Variant2Reg(TheValue,RegType,Buf,BufLen) then
        raise ERegFileError.Create('Invalid variant type');
     try
       RegPutData(Index,Buf,BufLen,RegType);
     finally
       FreeMem(Buf,BufLen);
       end;
   finally
     CloseKey;
     end;
end;

function TRegFile.BinaryDataSize(Index:String):integer;
var
  RegType,BufLen:DWORD;
  Buf:PByte;
begin
   OpenKey(fKey+ParseKey(Index),false);
   try
     BufLen := RegDataSize(Index,RegType);
     if (BufLen <= 0) or (RegType <> REG_BINARY) then
        Result := 0
     else
        Result := BufLen;
   finally
     CloseKey;
     end;
end;

procedure TRegFile.GetBinaryData(Index:String; var Buff; BuffMax:integer);
var
  RegType,BufLen:DWORD;
begin
   OpenKey(fKey+ParseKey(Index),false);
   try
     { Get the buffer size }
     BufLen := RegDataSize(Index,RegType);
     if (BufLen <= 0) or (RegType <> REG_BINARY) or (BufLen > BuffMax) then
        raise ERegFileError.Create('Value not compatible');
     { Now get the data }
     RegGetData(Index,@Buff,BufLen);
   finally
     CloseKey;
     end;
end;

procedure TRegFile.PutBinaryData(Index:String; var Buff; BuffMax:integer);
begin
   OpenKey(fKey+ParseKey(Index),false);
   try
     RegPutData(Index,@Buff,BuffMax,REG_BINARY);
   finally
     CloseKey;
     end;
end;

function TRegFile.GetValueDefault(Index:String; TheDefault:Variant):Variant;
begin
   Result := GetValue(Index);
   if VarIsNull(Result) then Result := TheDefault;
end;

procedure TRegFile.GetValueNames(ASubKey:String; AList:TStrings);
var
  RType,i,zTempLen : DWORD;
  zTemp : array[0..255] of char;
begin
  AList.Clear;
  if not OpenKey(MergeKey(ASubKey),true) then exit;
  try
    i := 0;
    zTempLen := 255;
    while RegEnumValue(fOpenKey,i,zTemp,zTempLen,nil,@RType,nil,nil) = ERROR_SUCCESS do
      begin
      AList.Add(StrPas(zTemp));
      zTempLen := 255;
      Inc(i);
      end;
  finally
   CloseKey;
   end;
end;

function TRegFile.HasSubKey(AKey:String):boolean;
var
  RType,i,zTempLen : DWORD;
  zTemp : array[0..255] of char;
  LastWriteTime:TFileTime;
begin
  Result := false;
  if not OpenKey(MergeKey(AKey),true) then exit;
  try
    i := 0;
    zTempLen := 255;
    Result := RegEnumKeyEx(fOpenKey,i,zTemp,zTempLen,
              nil,nil,nil,@LastWriteTime) = ERROR_SUCCESS;
  finally
   CloseKey;
   end;
end;

procedure TRegFile.GetSubKeys(AKey:String; AList:TStrings);
var
  RType,i,zTempLen : DWORD;
  zTemp : array[0..255] of char;
  LastWriteTime:TFileTime;
begin
  AList.Clear;
  if not OpenKey(MergeKey(AKey),true) then exit;
  try
    i := 0;
    zTempLen := 255;
    while RegEnumKeyEx(fOpenKey,i,zTemp,zTempLen,
           nil,nil,nil,@LastWriteTime) = ERROR_SUCCESS do
      begin
      AList.Add(StrPas(zTemp));
      zTempLen := 255;
      Inc(i);
      end;
  finally
   CloseKey;
   end;
end;

function TRegFile.MergeKey(AKey:String):String;
begin
  Result := fKey;
  if AKey <> '' then
     begin
     if Result <> '' then Result := Result+'\';
     Result := Result+AKey;
     end;
end;

function TRegFile.KeyExists;
begin
  Result := OpenKey(MergeKey(AKey),true);
  if Result then CloseKey;
end;

procedure TRegFile.DeleteKey(AKey:String);
begin
  if not OpenKey(fKey,true) then exit;
  RegCheck(RegDeleteKey(fOpenKey,@AKey[1]));
  CloseKey;
end;

procedure TRegFile.DeleteValue(AKey:String; AValue:String);
begin
  if not OpenKey(MergeKey(AKey)+ParseKey(AValue),true) then exit;
  RegCheck(RegDeleteValue(fOpenKey,PChar(AValue)));
  CloseKey;
end;

function TRegFile.RenameValue(AKey,OldName,NewName:String):boolean;
var
  RegType,BufLen:DWORD;
  Buf:PByte;
begin
  Result := false;
  if ValueExists(MergeKey(AKey),NewName) then exit;
  if not OpenKey(MergeKey(AKey),true) then exit;
  try
    BufLen := RegDataSize(OldName,RegType);
    if BufLen <= 0 then exit;
    Getmem(Buf,BufLen);
    try
      RegType := RegGetData(OldName,Buf,BufLen);
      RegCheck(RegDeleteValue(fOpenKey,PChar(OldName)));
      RegPutData(NewName,Buf,BufLen,RegType);
    finally
       FreeMem(Buf,BufLen);
       end;
  finally
     CloseKey;
     end;
  Result := true;
end;

function TRegFile.RenameKey(OldName,NewName:String):boolean;
var
  SrcKey,DestKey:HKEY;

  procedure CopyValues(Src,Dest:HKEY);
  var
    Buf:PByte;
    BufLen,RegType,RType:DWORD;
    i : integer;
    zTempLen: integer;
    zTemp: array[0..255] of char;
    Temp:String;
    SaveKey : HKEY;
  begin
    SaveKey := fOpenKey;
    try
    i := 0;
    zTempLen := 255;
    fOpenKey := Src;
    Temp := '';
    BufLen := RegDataSize(Temp,RegType);
    if BufLen > 0 then
       begin
       GetMem(Buf,BufLen);
       try
         RegType := RegGetData(Temp,Buf,Buflen);
         fOpenKey := Dest;
         RegPutData(Temp,Buf,BufLen,RegType);
       finally
         FreeMem(Buf,BufLen);
         end;
       end;
    zTempLen := 255;
    while RegEnumValue(Src,i,zTemp,zTempLen,nil,@RType,nil,nil) = ERROR_SUCCESS do
      begin
      fOpenKey := Src;
      Temp := StrPas(zTemp);
      BufLen := RegDataSize(Temp,RegType);
      if BufLen > 0 then
         begin
         GetMem(Buf,BufLen);
         try
           RegType := RegGetData(Temp,Buf,Buflen);
           fOpenKey := Dest;
           RegPutData(Temp,Buf,BufLen,RegType);
         finally
           FreeMem(Buf,BufLen);
           end;
         end;
      zTempLen := 255;
      Inc(i);
      end;
    finally
      fOpenKey := SaveKey;
      end;
  end;

  procedure ProcessSubKeys(Src,Dest:HKey);
  var
    SubSrc,SubDest:HKEY;
    lpdwDisposition:DWORD;
    i:integer;
    zTempLen:DWORD;
    zTemp:array[0..255] of char;
  begin
    i := 0;
    zTempLen := 255;
    while RegEnumKeyEx(Src,i,zTemp,zTempLen,
           nil,nil,nil,nil) = ERROR_SUCCESS do
     begin
     RegCheck(RegCreateKeyEx(
         Src,                // handle of an open key
         zTemp,              // address of subkey name
         0,                  // reserved
         PChar(''),          // address of class string
         0,                  // special options flag
         KEY_ALL_ACCESS,     // desired security access
         nil,                // address of key security structure
         SubSrc,             // address of buffer for opened handle
         @lpdwDisposition     // address of disposition value buffer
         ));
     RegCheck(RegCreateKeyEx(
         Dest,               // handle of an open key
         zTemp,              // address of subkey name
         0,                  // reserved
         PChar(''),          // address of class string
         0,                  // special options flag
         KEY_ALL_ACCESS,     // desired security access
         nil,                // address of key security structure
         SubDest,            // address of buffer for opened handle
         @lpdwDisposition     // address of disposition value buffer
         ));
     CopyValues(SubSrc,SubDest);
     ProcessSubKeys(SubSrc,SubDest);
     zTempLen := 255;
     Inc(i);
     end;
  RegCloseKey(Src);
  RegCloseKey(Dest);
  end;

begin
  Result := true;
  if not OpenKey(MergeKey(OldName),true) then exit;
  SrcKey := fOpenKey;
  if not OpenKey(MergeKey(NewName),false) then
     begin
     RegCloseKey(SrcKey);
     exit;
     end;
  DestKey := fOpenKey;
  CopyValues(SrcKey,DestKey);
  ProcessSubKeys(SrcKey,DestKey);
  DeleteKey(OldName);
end;

end.



